"""Logging and log related utilities."""

from __future__ import annotations

from .formatters import console_log_formatter, json_formatter, key_value_formatter
from .job_logging_service import (
    JobLoggingService,
    MissingJobLogException,
    SizeThresholdJobLogException,
)
from .output_logger import OutputLogger
from .utils import DEFAULT_LEVEL, LEVELS, capture_subprocess_output, setup_logging

__all__ = [
    "JobLoggingService",
    "MissingJobLogException",
    "SizeThresholdJobLogException",
    "OutputLogger",
    "DEFAULT_LEVEL",
    "LEVELS",
    "capture_subprocess_output",
    "setup_logging",
    "console_log_formatter",
    "json_formatter",
    "key_value_formatter",
]



"""`SingerBlock` wraps singer plugins to implement the `IOBlock` interface."""

from __future__ import annotations

import asyncio
import typing as t
from contextlib import suppress

from meltano.core.block.ioblock import IOBlock
from meltano.core.logging import capture_subprocess_output
from meltano.core.plugin import PluginType
from meltano.core.runner import RunnerError

if t.TYPE_CHECKING:
    from asyncio.subprocess import Process

    from meltano.core.elt_context import PluginContext
    from meltano.core.logging.utils import SubprocessOutputWriter
    from meltano.core.plugin_invoker import PluginInvoker
    from meltano.core.project import Project

PRODUCERS = (PluginType.EXTRACTORS, PluginType.MAPPERS)
CONSUMERS = (PluginType.LOADERS, PluginType.MAPPERS)


class IOLinkError(Exception):
    """Raised when an IO link is not possible."""


class ProcessWaitError(Exception):
    """Raised when a process can be waited on."""


class InvokerBase:  # noqa: WPS230, WPS214
    """Base class for creating IOBlock's built on top of existing Meltano plugins."""

    def __init__(
        self,
        block_ctx: PluginContext,
        project: Project,
        plugin_invoker: PluginInvoker,
        command: str | None,
    ):
        """Init a wrapped plugin invoker for use as an IOBlock or PluginCommandBlock.

        Args:
            block_ctx: Context that should be used for this instance to do
                things like obtaining project settings.
            project: The Meltano project this `InvokerBase` is for.
            plugin_invoker: The actual plugin invoker.
            command: The optional command to invoke.
        """
        self.context = block_ctx
        self.project = project

        self.invoker: PluginInvoker = plugin_invoker
        self._command: str | None = command

        self.outputs = []
        self.err_outputs = []

        self.process_handle: Process | None = None
        self._process_future: asyncio.Task | None = None
        self._stdout_future: asyncio.Task | None = None
        self._stderr_future: asyncio.Task | None = None

    @property
    def command(self) -> str | None:
        """Get the command to use when invoking the plugin.

        Returns:
            The command to use when invoking the plugin.
        """
        return self._command

    @property
    def string_id(self) -> str:
        """Return a string identifier for this block.

        Returns:
            A string identifier for this block.
        """
        return self.invoker.plugin.name

    async def start(self, *args, **kwargs):
        """Invoke the process asynchronously.

        Args:
            args: arguments to pass to the process invoker.
            kwargs: keyword arguments to pass to the process invoker.

        Raises:
            RunnerError: If the plugin process can not start.
        """
        try:
            self.process_handle = await self.invoker.invoke_async(
                *args,
                command=self.command,
                **kwargs,
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.PIPE,
            )
        except Exception as err:
            raise RunnerError(f"Cannot start plugin: {err}") from err  # noqa: EM102

    async def stop(self, kill: bool = True):
        """Stop (kill) the underlying process and cancel output proxying.

        Args:
            kill: whether to send a SIGKILL. If false, a SIGTERM is sent.
        """
        if self.process_handle is None:
            return

        with suppress(ProcessLookupError):
            if kill:
                self.process_handle.kill()
            else:
                self.process_handle.terminate()
        await self.process_future
        if self._stdout_future is not None:
            self._stdout_future.cancel()
        if self._stderr_future is not None:
            self._stderr_future.cancel()

    def proxy_stdout(self) -> asyncio.Task:
        """Start proxying stdout to the linked stdout destinations.

        Returns:
            The stdout proxy future.

        Raises:
            IOLinkError: If the processes is not running and so there is no IO
                to proxy.
        """
        if self.process_handle is None:
            raise IOLinkError("No IO to proxy, process not running")  # noqa: EM101

        if self._stdout_future is None:
            outputs = self._merge_outputs(self.invoker.StdioSource.STDOUT, self.outputs)
            self._stdout_future = asyncio.ensure_future(
                # forward subproc stdout to downstream (i.e. targets stdin, loggers)
                capture_subprocess_output(self.process_handle.stdout, *outputs),
            )
        return self._stdout_future

    def proxy_stderr(self) -> asyncio.Task:
        """Start proxying stderr to the linked stderr destinations.

        Returns:
            The stderr proxy future.

        Raises:
            IOLinkError: If the processes is not running and so there is no IO
                to proxy.
        """
        if self.process_handle is None:
            raise IOLinkError("No IO to proxy, process not running")  # noqa: EM101

        if self._stderr_future is None:
            err_outputs = self._merge_outputs(
                self.invoker.StdioSource.STDERR,
                self.err_outputs,
            )
            self._stderr_future = asyncio.ensure_future(
                capture_subprocess_output(self.process_handle.stderr, *err_outputs),
            )
        return self._stderr_future

    def proxy_io(self) -> tuple[asyncio.Task, asyncio.Task]:
        """Start proxying stdout AND stderr to the respectively linked destinations.

        Returns:
            proxy_stdout asyncio.Task and proxy_stderr asyncio.Task
        """
        return self.proxy_stdout(), self.proxy_stderr()

    @property
    def process_future(self) -> asyncio.Task:
        """Return the future of the underlying process wait() call.

        Returns:
            The future of the underlying process wait() calls.

        Raises:
            ProcessWaitError: If the process is not running.
        """
        if self._process_future is None:
            if self.process_handle is None:
                raise ProcessWaitError(
                    "No process to wait, process not running running",  # noqa: EM101
                )
            self._process_future = asyncio.ensure_future(self.process_handle.wait())
        return self._process_future

    @property
    def stdin(self) -> asyncio.StreamWriter | None:
        """Return stdin of the underlying process.

        Returns:
            The stdin of the underlying process.
        """
        return None if self.process_handle is None else self.process_handle.stdin

    async def close_stdin(self) -> None:
        """Close the underlying process stdin if the block is a consumer."""
        if self.consumer:
            self.process_handle.stdin.close()
            await self.process_handle.stdin.wait_closed()

    def stdout_link(self, dst: SubprocessOutputWriter) -> None:
        """Use stdout_link to instruct block to link/write stdout content to dst.

        Args:
            dst:  The destination stdout output should be written too.

        Raises:
            IOLinkError: If the IO is already in flight.
        """
        if self._stdout_future is None:
            self.outputs.append(dst)
        else:
            raise IOLinkError("IO capture already in flight")  # noqa: EM101

    def stderr_link(self, dst: SubprocessOutputWriter):
        """Use stderr_link to instruct block to link/write stderr content to dst.

        Args:
            dst:  The destination stderr output should be written too.

        Raises:
            IOLinkError: If the IO is already in flight.
        """
        if self._stderr_future is None:
            self.err_outputs.append(dst)
        else:
            raise IOLinkError("IO capture already in flight")  # noqa: EM101

    async def pre(self, context) -> None:
        """Pre triggers preparation of the underlying plugin.

        Args:
            context: The context with which to update the invoker
        """
        self.invoker.context = context
        await self.invoker.prepare(context.session)

    async def post(self) -> None:
        """Post triggers resetting the underlying plugin config."""
        with suppress(FileNotFoundError):
            # TODO: should we preserve these on a failure?
            # The invoker prepared context manager was able to clean up the configs
            await self.invoker.cleanup()

    def _merge_outputs(self, source: str, outputs: list) -> list:
        if not self.invoker.output_handlers:
            return outputs

        merged_outputs = self.invoker.output_handlers.get(source, [])
        merged_outputs.extend(outputs)
        return merged_outputs


class SingerBlock(InvokerBase, IOBlock):
    """SingerBlock wraps singer plugins to implement the IOBlock interface."""

    def __init__(
        self,
        block_ctx: PluginContext,
        project: Project,
        plugin_invoker: PluginInvoker,
        plugin_args: t.Sequence[str],
    ):
        """Configure and return a Singer plugin wrapped as an IOBlock.

        Args:
            block_ctx: the block context.
            project:  the project to use to obtain project settings.
            plugin_invoker: the plugin invoker.
            plugin_args: any additional plugin args that should be used.
        """
        super().__init__(
            block_ctx=block_ctx,
            project=project,
            plugin_invoker=plugin_invoker,
            command=None,
        )
        self.plugin_args = plugin_args

    @property
    def producer(self) -> bool:
        """Whether this plugin is a producer.

        Currently if the underlying plugin is of type extractor, it is a producer.

        Returns:
            True if the underlying plugin is a producer, False otherwise.
        """
        return self.invoker.plugin.type in PRODUCERS

    @property
    def consumer(self) -> bool:
        """Whether this plugin is a consumer.

        Currently if the underlying plugin is of type loader, it is a consumer.

        Returns:
            True if the plugin is a consumer, False otherwise.
        """
        return self.invoker.plugin.type in CONSUMERS

    @property
    def has_state(self) -> bool:
        """Whether this plugin has state.

        Returns:
            bool indicating whether this plugin has state
        """
        return "state" in self.invoker.capabilities

    async def start(self):
        """Start the SingerBlock by invoking the underlying plugin.

        Raises:
            RunnerError: If the plugin can not start.
        """
        stream_buffer_size = self.project.settings.get("elt.buffer_size")
        line_length_limit = stream_buffer_size // 2

        stdin = asyncio.subprocess.PIPE if self.consumer else None
        try:
            self.process_handle = await self.invoker.invoke_async(
                limit=line_length_limit,
                stdin=stdin,  # Singer messages
                stdout=asyncio.subprocess.PIPE,  # Singer state
                stderr=asyncio.subprocess.PIPE,  # Log
            )
        except Exception as err:
            raise RunnerError(f"Cannot start plugin {self.string_id}: {err}") from err  # noqa: EM102

    async def stop(self, kill: bool = True):
        """Stop (kill) the underlying process and cancel output proxying.

        Args:
            kill: Whether to send a SIGKILL. If false, a SIGTERM is sent.
        """
        if self.process_handle is None:
            return

        with suppress(ProcessLookupError):
            if kill:
                self.process_handle.kill()
            else:
                self.process_handle.terminate()
        await self.process_future
        if self._stdout_future is not None:
            self._stdout_future.cancel()
        if self._stderr_future is not None:
            self._stderr_future.cancel()
        with suppress(FileNotFoundError):
            # the invoker prepared context manager was able to clean up the configs
            await self.invoker.cleanup()



from __future__ import annotations

import typing as t

import pytest
import yaml
from mock import Mock
from sqlalchemy.exc import OperationalError

from meltano.core.db import NullConnectionStringError, connect, project_engine
from meltano.core.project import Project

if t.TYPE_CHECKING:
    from pathlib import Path


class TestConnectionRetries:
    def test_ping_failure(self):
        engine_mock = Mock()

        # check if OperationalError is raised if a connection can't be made
        engine_mock.connect.side_effect = OperationalError(
            "test_error",
            "test_error",
            "test_error",
        )
        with pytest.raises(OperationalError):
            connect(engine=engine_mock, max_retries=3, retry_timeout=0.1)

        assert engine_mock.connect.call_count == 4

        # check reconnect on second call to `engine.connect`
        engine_mock.reset_mock()
        engine_mock.connect.side_effect = [
            OperationalError("test_error", "test_error", "test_error"),
            None,
        ]

        connect(engine=engine_mock, max_retries=3, retry_timeout=0.1)
        assert engine_mock.connect.call_count == 2


class TestProjectEngine:
    def test_project_engine(
        self,
        monkeypatch: pytest.MonkeyPatch,
        tmp_path: Path,
    ):
        with tmp_path.joinpath("meltano.yml").open("w") as meltano_yml:
            yaml.dump(
                {
                    "project_id": "test",
                    "send_anonymous_usage_stat": False,
                    "database_uri": "$DATABASE",
                },
                meltano_yml,
            )
        monkeypatch.delenv("MELTANO_DATABASE_URI")

        project = Project(tmp_path)
        with pytest.raises(NullConnectionStringError):
            project_engine(project)



from __future__ import annotations

import mock
import pytest

from asserts import assert_cli_runner
from meltano.cli import cli
from meltano.core.plugin import PluginType
from meltano.core.project_add_service import PluginAlreadyAddedException


class TestCliRemove:
    @pytest.fixture(scope="class")
    def tap_gitlab(self, project_add_service):
        try:
            return project_add_service.add(PluginType.EXTRACTORS, "tap-gitlab")
        except PluginAlreadyAddedException as err:
            return err.plugin

    def test_remove(self, project, tap, cli_runner):
        with mock.patch("meltano.cli.remove.remove_plugins") as remove_plugins_mock:
            result = cli_runner.invoke(cli, ["remove", tap.type, tap.name])
            assert_cli_runner(result)

            remove_plugins_mock.assert_called_once_with(project, [tap])

    def test_remove_multiple(self, project, tap, tap_gitlab, cli_runner):
        with mock.patch("meltano.cli.remove.remove_plugins") as remove_plugins_mock:
            result = cli_runner.invoke(
                cli,
                ["remove", "extractors", tap.name, tap_gitlab.name],
            )
            assert_cli_runner(result)

            remove_plugins_mock.assert_called_once_with(project, [tap, tap_gitlab])

    def test_remove_type_name(self, project, tap, target, cli_runner):
        with mock.patch("meltano.cli.remove.remove_plugins") as remove_plugins_mock:
            result = cli_runner.invoke(cli, ["remove", "extractor", tap.name])
            assert_cli_runner(result)

            remove_plugins_mock.assert_called_with(project, [tap])

            result = cli_runner.invoke(cli, ["remove", "loader", target.name])
            assert_cli_runner(result)

            remove_plugins_mock.assert_called_with(project, [target])

            assert remove_plugins_mock.call_count == 2



from __future__ import annotations

import json
import typing as t

import pytest

from meltano.core.plugin import PluginType

if t.TYPE_CHECKING:
    from sqlalchemy.orm import Session

    from meltano.core.plugin.project_plugin import ProjectPlugin
    from meltano.core.plugin_invoker import PluginInvoker
    from meltano.core.project_add_service import ProjectAddService


class TestSingerMapper:
    @pytest.fixture(scope="class")
    def subject(self, project_add_service: ProjectAddService) -> ProjectPlugin:
        return project_add_service.add(
            PluginType.MAPPERS,
            "mapper-mock",
            mappings=[
                {
                    "name": "mock-mapping-0",
                    "config": {
                        "transformations": [
                            {
                                "field_id": "author_email",
                                "tap_stream_name": "commits",
                                "type": "MASK-HIDDEN",
                            },
                        ],
                    },
                },
                {
                    "name": "mock-mapping-1",
                    "config": {
                        "transformations": [
                            {
                                "field_id": "given_name",
                                "tap_stream_name": "users",
                                "type": "lowercase",
                            },
                        ],
                    },
                },
            ],
            config={
                "parent_mapper_setting": "parent_mapper_setting_value",
                "transformations": [
                    {
                        "field_id": "given_name",
                        "tap_stream_name": "users",
                        "type": "uppercase",
                    },
                ],
            },
            mapping_name="mock-mapping-0",
        )

    @pytest.mark.asyncio()
    async def test_exec_args(
        self,
        subject: ProjectPlugin,
        session: Session,
        plugin_invoker_factory: t.Callable[[ProjectPlugin], PluginInvoker],
    ):
        invoker = plugin_invoker_factory(subject)
        async with invoker.prepared(session):
            assert subject.exec_args(invoker) == ["--config", invoker.files["config"]]

    @pytest.mark.asyncio()
    async def test_merged_config(
        self,
        subject: ProjectPlugin,
        session: Session,
        plugin_invoker_factory: t.Callable[[ProjectPlugin], PluginInvoker],
    ):
        invoker = plugin_invoker_factory(subject)

        async with invoker.prepared(session):
            config_path = invoker.files["config"]

            with open(config_path) as config_file:
                config = json.load(config_file)

            assert config == {
                "transformations": [
                    {
                        "field_id": "author_email",
                        "tap_stream_name": "commits",
                        "type": "MASK-HIDDEN",
                    }
                ],
                "parent_mapper_setting": "parent_mapper_setting_value",
            }



"""Plugin requirements."""

from __future__ import annotations

import typing as t

from meltano.core.behavior import NameEq
from meltano.core.behavior.canonical import Canonical

TReq = t.TypeVar("TReq", bound="PluginRequirement")


class PluginRequirement(NameEq, Canonical):
    """A plugin requirement."""

    def __init__(self, name: str, variant: str | None = None) -> None:
        """Create a new PluginBundleFile.

        Args:
            name: The name of the plugin dependency.
            variant: The variant of the plugin dependency.
        """
        super().__init__(name=name, variant=variant)

    @classmethod
    def parse_all(cls: type[TReq], obj: dict | None) -> dict[str, list[TReq]]:
        """Deserialize requirements data into a dict of PluginRequirement.

        Args:
            obj: Raw Python object.

        Returns:
            Mapping of plugin types to requirements.
        """
        if obj is not None:
            return {
                plugin_type: [cls.parse(req) for req in requirements]
                for plugin_type, requirements in obj.items()
            }

        return {}



from __future__ import annotations

import platform
import typing as t
from datetime import datetime, timezone

import dotenv
import pytest

from meltano.core.plugin import PluginType
from meltano.core.plugin.project_plugin import ProjectPlugin
from meltano.core.project_plugins_service import PluginAlreadyAddedException
from meltano.core.setting import Setting
from meltano.core.settings_service import (
    FEATURE_FLAG_PREFIX,
    REDACTED_VALUE,
    FeatureFlags,
    SettingValueStore,
)
from meltano.core.settings_store import (
    ConflictingSettingValueException,
    MultipleEnvVarsSetException,
)
from meltano.core.utils import EnvironmentVariableNotSetError

if t.TYPE_CHECKING:
    from meltano.core.environment import Environment
    from meltano.core.plugin.settings_service import PluginSettingsService
    from meltano.core.project import Project


@pytest.mark.order(0)
def test_create(session):
    setting = Setting(
        name="api_key.test.test",
        namespace="gitlab",
        value="C4F3C4F3",
        enabled=True,
    )

    session.add(setting)
    session.commit()

    fetched = session.query(Setting).first()
    assert setting == fetched


@pytest.fixture(scope="class")
def env_var():
    def _wrapper(plugin_settings_service, setting_name):
        setting_def = plugin_settings_service.find_setting(setting_name)
        return plugin_settings_service.setting_env(setting_def)

    return _wrapper


@pytest.fixture(scope="class")
def custom_tap(project):
    expected = {"test": "custom", "start_date": None, "secure": None}
    tap = ProjectPlugin(
        PluginType.EXTRACTORS,
        name="tap-custom",
        namespace="tap_custom",
        config=expected,
    )
    try:
        return project.plugins.add_to_file(tap)
    except PluginAlreadyAddedException as err:
        return err.plugin


@pytest.fixture()
def subject(tap, plugin_settings_service_factory) -> PluginSettingsService:
    return plugin_settings_service_factory(tap)


@pytest.fixture()
def environment(project: Project) -> Environment:
    project.activate_environment("dev")
    try:
        yield project.environment
    finally:
        project.deactivate_environment()


class TestPluginSettingsService:
    def test_get_with_source(
        self,
        session,
        tap,
        inherited_tap,
        env_var,
        monkeypatch,
        plugin_settings_service_factory,
    ):
        subject = plugin_settings_service_factory(inherited_tap)

        # returns the default value when unset
        assert subject.get_with_source("test", session=session) == (
            "mock",
            SettingValueStore.DEFAULT,
        )

        # returns the inherited value when set
        parent_subject = plugin_settings_service_factory(tap)
        monkeypatch.setenv(env_var(parent_subject, "test"), "INHERITED")

        assert subject.get_with_source("test", session=session) == (
            "INHERITED",
            SettingValueStore.INHERITED,
        )

        # overriden by an Setting db value when set
        subject.set(
            "test",
            "THIS_IS_FROM_DB",
            store=SettingValueStore.DB,
            session=session,
        )

        assert subject.get_with_source("test", session=session) == (
            "THIS_IS_FROM_DB",
            SettingValueStore.DB,
        )

        # overriden via the `meltano.yml` configuration
        test_value = 42
        subject.set(
            "test",
            test_value,
            store=SettingValueStore.MELTANO_YML,
            session=session,
        )

        assert subject.get_with_source("test", session=session) == (
            test_value,
            SettingValueStore.MELTANO_YML,
        )

        # revert back to the original
        subject.reset(store=SettingValueStore.MELTANO_YML)

        # overriden via ENV
        monkeypatch.setenv(env_var(subject, "test"), "N33DC0F33")

        assert subject.get_with_source("test", session=session) == (
            "N33DC0F33",
            SettingValueStore.ENV,
        )

        # overridden via config override
        monkeypatch.setitem(subject.config_override, "test", "foo")

        assert subject.get_with_source("test", session=session) == (
            "foo",
            SettingValueStore.CONFIG_OVERRIDE,
        )

    def test_get_with_source_casting(self, session, subject, env_var, monkeypatch):
        # Verify that integer settings set in env are cast correctly
        monkeypatch.setenv(env_var(subject, "port"), "3333")

        assert subject.get_with_source("port", session=session) == (
            3333,
            SettingValueStore.ENV,
        )

        # Verify that array settings set in env are cast correctly
        monkeypatch.setenv(env_var(subject, "list"), '[1, 2, 3, "4"]')

        assert subject.get_with_source("list", session=session) == (
            [1, 2, 3, "4"],
            SettingValueStore.ENV,
        )

        # Verify that object settings set in env are cast correctly
        monkeypatch.setenv(env_var(subject, "object"), '{"1":{"2":3}}')

        assert subject.get_with_source("object", session=session) == (
            {"1": {"2": 3}},
            SettingValueStore.ENV,
        )

        # Verify that boolean settings set in env are cast correctly
        # Default
        assert subject.get_with_source("boolean", session=session) == (
            None,
            SettingValueStore.DEFAULT,
        )

        # Preferred env var
        monkeypatch.setenv(env_var(subject, "boolean"), "0")

        assert subject.get_with_source("boolean", session=session) == (
            False,
            SettingValueStore.ENV,
        )

    def test_definitions(self, subject):
        subject.show_hidden = False
        subject._setting_defs = None

        setting_defs_by_name = {sdef.name: sdef for sdef in subject.definitions()}

        # Regular settings
        assert "test" in setting_defs_by_name
        assert "start_date" in setting_defs_by_name

        # Expect hidden
        assert "secret" not in setting_defs_by_name

    @pytest.mark.order(1)
    @pytest.mark.usefixtures("tap")
    def test_as_dict(self, subject, session):
        expected = {"test": "mock", "start_date": None, "secure": None}
        full_config = subject.as_dict(session=session)
        redacted_config = subject.as_dict(redacted=True, session=session)

        for key, value in expected.items():
            assert full_config.get(key) == value
            assert redacted_config.get(key) == value

    def test_environment_only_config(
        self,
        environment: Environment,
        subject: PluginSettingsService,
        project: Project,
    ):
        subject.set("test_environment", "THIS_IS_FROM_ENVIRONMENT")
        by_name = {sdef.name: sdef for sdef in subject.definitions()}

        assert "test_environment" in by_name
        assert not by_name["test"].is_custom
        assert by_name["test_environment"].is_custom

        with project.meltano_update() as meltano:
            plugin = meltano.plugins.extractors[0]
            assert "test_environment" not in plugin.config

            dev_environment = meltano.environments[0]
            env_plugin = dev_environment.config.plugins[PluginType.EXTRACTORS][0]
            assert dev_environment == environment
            assert env_plugin.config["test_environment"] == "THIS_IS_FROM_ENVIRONMENT"

    @pytest.mark.usefixtures("tap")
    def test_as_dict_process(self, subject: PluginSettingsService):
        config = subject.as_dict()
        assert config["auth.username"] is None
        assert config["auth.password"] is None
        assert "auth" not in config

        config = subject.as_dict(process=True)
        assert "auth.username" not in config
        assert "auth.password" not in config
        assert "auth" not in config

        subject.set("auth.username", "nested_username")

        config = subject.as_dict()
        assert config["auth.username"] == "nested_username"
        assert config["auth.password"] is None
        assert "auth" not in config

        config = subject.as_dict(process=True)
        assert config["auth"]["username"] == "nested_username"
        assert "password" not in config["auth"]
        assert "auth.username" not in config
        assert "auth.password" not in config

        subject.set("auth.password", "nested_password")

        config = subject.as_dict()
        assert config["auth.username"] == "nested_username"
        assert config["auth.password"] == "nested_password"
        assert "auth" not in config

        config = subject.as_dict(process=True)
        assert config["auth"]["username"] == "nested_username"
        assert config["auth"]["password"] == "nested_password"  # noqa: S105
        assert "auth.username" not in config
        assert "auth.password" not in config

    @pytest.mark.usefixtures("project")
    def test_as_dict_custom(
        self,
        session,
        custom_tap,
        plugin_settings_service_factory,
    ):
        subject = plugin_settings_service_factory(custom_tap)
        assert subject.as_dict(extras=False, session=session) == custom_tap.config

    @pytest.mark.usefixtures("tap")
    def test_as_dict_redacted(self, subject, session):
        store = SettingValueStore.DB

        # ensure values are redacted when they are set
        subject.set("secure", "thisisatest", store=store, session=session)

        config = subject.as_dict(redacted=True, session=session)
        assert config["secure"] == REDACTED_VALUE

        config = subject.as_dict(redacted=True, session=session, redacted_value="*****")
        assert config["secure"] == "*****"

        # although setting the REDACTED_VALUE does nothing
        subject.set("secure", REDACTED_VALUE, store=store, session=session)
        config = subject.as_dict(session=session)
        assert config["secure"] == "thisisatest"

    @pytest.mark.usefixtures("tap")
    def test_as_env(self, subject, session, env_var):
        subject.set("boolean", True, store=SettingValueStore.DOTENV)
        subject.set("list", [1, 2, 3, "4"], store=SettingValueStore.DOTENV)
        subject.set("object", {"1": {"2": 3}}, store=SettingValueStore.DOTENV)

        config = subject.as_env(session=session)
        subject.reset(store=SettingValueStore.DOTENV)

        # Settings with values are present
        assert config[env_var(subject, "test")] == "mock"
        assert config[env_var(subject, "list")] == '[1, 2, 3, "4"]'
        assert config[env_var(subject, "object")] == '{"1": {"2": 3}}'
        assert config[env_var(subject, "boolean")] == "true"

        # Settings without values are not
        assert env_var(subject, "start_date") not in config
        assert env_var(subject, "secure") not in config

        # Generic env vars are present
        assert config["MELTANO_EXTRACT_TEST"] == "mock"
        assert config["MELTANO_EXTRACT_LIST"] == '[1, 2, 3, "4"]'
        assert config["MELTANO_EXTRACT_OBJECT"] == '{"1": {"2": 3}}'
        assert config["MELTANO_EXTRACT_BOOLEAN"] == "true"

    @pytest.mark.usefixtures("project")
    def test_as_env_custom(
        self,
        session,
        custom_tap,
        env_var,
        plugin_settings_service_factory,
    ):
        subject = plugin_settings_service_factory(custom_tap)
        config = subject.as_env(session=session)
        for key, value in custom_tap.config.items():
            assert config.get(env_var(subject, key)) == value

    @pytest.mark.order(4)
    @pytest.mark.usefixtures("env_var")
    def test_namespace_as_env_prefix(
        self,
        project: Project,
        session,
        target: ProjectPlugin,
        plugin_settings_service_factory,
    ):
        subject = plugin_settings_service_factory(target)

        def assert_env_value(value, env_var):
            read_value, metadata = subject.get_with_metadata("schema")
            assert value == read_value
            assert metadata["env_var"] == env_var

        assert subject.get("schema") is None

        subject.set("schema", "default", store=SettingValueStore.DOTENV)
        value, metadata = subject.get_with_metadata("schema")

        # Env is the default
        assert_env_value("default", "TARGET_MOCK_SCHEMA")

        subject.unset("schema")

        # Namespace prefix
        dotenv.set_key(project.dotenv, "MOCK_SCHEMA", "namespace_prefix")
        assert_env_value("namespace_prefix", "MOCK_SCHEMA")

        # Name prefix
        dotenv.unset_key(project.dotenv, "MOCK_SCHEMA")
        dotenv.set_key(project.dotenv, "TARGET_MOCK_SCHEMA", "name_prefix")
        project.refresh()
        assert_env_value("name_prefix", "TARGET_MOCK_SCHEMA")

        config = subject.as_env(session=session)
        subject.reset(store=SettingValueStore.DOTENV)

        assert (
            config["MELTANO_LOAD_SCHEMA"] == "name_prefix"
        )  # Generic prefix, read-only

    def test_setting_env_vars(
        self,
        tap,
        inherited_tap,
        alternative_target,
        plugin_settings_service_factory,
    ):
        def env_vars(service, setting_name, **kwargs):
            return [
                setting.definition
                for setting in service.setting_env_vars(
                    service.find_setting(setting_name),
                    **kwargs,
                )
            ]

        # Shadowing base plugin
        service = plugin_settings_service_factory(tap)
        # For reading setting values from environment
        assert env_vars(service, "boolean") == [
            "TAP_MOCK_BOOLEAN",  # Name and namespace prefix
        ]
        # For writing values into the execution environment
        assert env_vars(service, "boolean", for_writing=True) == [
            "TAP_MOCK_BOOLEAN",  # Name and namespace prefix
            "MELTANO_EXTRACT_BOOLEAN",  # Generic prefix
        ]

        # Inheriting from base plugin
        service = plugin_settings_service_factory(alternative_target)
        # For reading setting values from environment
        assert env_vars(service, "schema") == [
            "TARGET_MOCK_ALTERNATIVE_SCHEMA",  # Name and namespace prefix
        ]
        # For writing values into the execution environment
        assert env_vars(service, "schema", for_writing=True) == [
            "MOCKED_SCHEMA",  # Custom `env`
            "TARGET_MOCK_ALTERNATIVE_SCHEMA",  # Name and namespace prefix
            "TARGET_MOCK_SCHEMA",  # Parent name prefix
            "MOCK_SCHEMA",  # Parent namespace prefix
            "MELTANO_LOAD_SCHEMA",  # Generic prefix
        ]

        # Inheriting from project plugin
        service = plugin_settings_service_factory(inherited_tap)
        # For reading setting values from environment
        assert env_vars(service, "boolean") == [
            "TAP_MOCK_INHERITED_BOOLEAN",  # Name and namespace prefix
        ]
        # For writing values into the execution environment
        assert env_vars(service, "boolean", for_writing=True) == [
            "TAP_MOCK_INHERITED_BOOLEAN",  # Name and namespace prefix
            "TAP_MOCK_BOOLEAN",  # Parent name and namespace prefix
            "MELTANO_EXTRACT_BOOLEAN",  # Generic prefix
        ]

    @pytest.mark.usefixtures("tap")
    def test_store_db(self, session, subject):
        store = SettingValueStore.DB

        subject.set("test_a", "THIS_IS_FROM_DB", store=store, session=session)
        subject.set("test_b", "THIS_IS_FROM_DB", store=store, session=session)

        assert session.query(Setting).count() == 2

        subject.unset("test_a", store=store, session=session)

        assert session.query(Setting).count() == 1

        subject.reset(store=store, session=session)

        assert session.query(Setting).count() == 0

    @pytest.mark.usefixtures("tap")
    def test_store_meltano_yml(self, subject, project):
        store = SettingValueStore.MELTANO_YML

        subject.set("test_a", "THIS_IS_FROM_YML", store=store)
        subject.set("test_b", "THIS_IS_FROM_YML", store=store)

        with project.meltano_update() as meltano:
            extractor = meltano.plugins.extractors[0]
            assert extractor.config["test_a"] == "THIS_IS_FROM_YML"
            assert extractor.config["test_b"] == "THIS_IS_FROM_YML"

        subject.unset("test_a", store=store)

        with project.meltano_update() as meltano:
            extractor = meltano.plugins.extractors[0]
            assert "test_a" not in extractor.config
            assert extractor.config["test_b"] == "THIS_IS_FROM_YML"

        subject.reset(store=store)

        with project.meltano_update() as meltano:
            extractor = meltano.plugins.extractors[0]
            assert "test_a" not in extractor.config
            assert "test_b" not in extractor.config

    @pytest.mark.order(2)
    @pytest.mark.usefixtures("tap")
    def test_store_dotenv(self, subject: PluginSettingsService, project: Project):
        store = SettingValueStore.DOTENV

        assert not project.dotenv.exists()

        subject.set("test", "THIS_IS_FROM_DOTENV", store=store)
        subject.set("start_date", "THIS_IS_FROM_DOTENV", store=store)

        dotenv_contents = dotenv.dotenv_values(project.dotenv)
        assert dotenv_contents["TAP_MOCK_TEST"] == "THIS_IS_FROM_DOTENV"
        assert dotenv_contents["TAP_MOCK_START_DATE"] == "THIS_IS_FROM_DOTENV"
        assert subject.get_with_source("test") == (
            "THIS_IS_FROM_DOTENV",
            SettingValueStore.DOTENV,
        )
        assert subject.get_with_source("start_date") == (
            "THIS_IS_FROM_DOTENV",
            SettingValueStore.DOTENV,
        )

        dotenv.set_key(project.dotenv, "TAP_MOCK_BOOLEAN", "false")
        project.refresh()
        assert subject.get_with_source("boolean") == (False, SettingValueStore.DOTENV)
        dotenv.unset_key(project.dotenv, "TAP_MOCK_BOOLEAN")

        subject.set("boolean", True, store=store)

        dotenv_contents = dotenv.dotenv_values(project.dotenv)
        assert dotenv_contents["TAP_MOCK_BOOLEAN"] == "true"
        assert subject.get_with_source("boolean") == (True, SettingValueStore.DOTENV)

        subject.set("list", [1, 2, 3, "4"], store=store)

        dotenv_contents = dotenv.dotenv_values(project.dotenv)
        assert dotenv_contents["TAP_MOCK_LIST"] == '[1, 2, 3, "4"]'
        assert subject.get_with_source("list") == (
            [1, 2, 3, "4"],
            SettingValueStore.DOTENV,
        )

        subject.set("object", {"1": {"2": 3}}, store=store)

        dotenv_contents = dotenv.dotenv_values(project.dotenv)
        assert dotenv_contents["TAP_MOCK_OBJECT"] == '{"1": {"2": 3}}'
        assert subject.get_with_source("object") == (
            {"1": {"2": 3}},
            SettingValueStore.DOTENV,
        )

        subject.unset("test", store=store)

        dotenv_contents = dotenv.dotenv_values(project.dotenv)
        assert "TAP_MOCK_TEST" not in dotenv_contents
        assert dotenv_contents["TAP_MOCK_START_DATE"] == "THIS_IS_FROM_DOTENV"
        assert dotenv_contents["TAP_MOCK_BOOLEAN"] == "true"

        subject.reset(store=store)
        assert not project.dotenv.exists()

    @pytest.mark.usefixtures("tap")
    def test_env_var_expansion(
        self,
        session,
        subject: PluginSettingsService,
        project: Project,
        monkeypatch,
        env_var,
    ):
        if platform.system() == "Windows":
            pytest.xfail(
                "Fails on Windows: https://github.com/meltano/meltano/issues/3444",
            )
        monkeypatch.setenv("VAR", "hello world!")
        monkeypatch.setenv("FOO", "42")

        project.dotenv.touch()
        dotenv.set_key(project.dotenv, "A", "rock")
        dotenv.set_key(project.dotenv, "B", "paper")
        dotenv.set_key(project.dotenv, "C", "scissors")
        project.refresh()

        yml_config = {
            "var": "$VAR",
            "foo": "${FOO}",
            "missing": "$MISSING",
            "multiple": "$A ${B} $C",
            "info": "$MELTANO_EXTRACTOR_NAME",
            "password": "foo$r$6$bar",
            "_extra": "$TAP_MOCK_MULTIPLE",
            "_extra_generic": "$MELTANO_EXTRACT_FOO",
        }
        monkeypatch.setattr(subject.plugin, "config", yml_config)

        # Env vars inside env var values do not get expanded
        monkeypatch.setenv(env_var(subject, "test"), "$FOO")

        config = subject.as_dict(session=session)

        assert config["var"] == "hello world!"
        assert config["foo"] == "42"
        assert config["missing"] is None
        assert config["multiple"] == "rock paper scissors"
        assert config["info"] == "tap-mock"

        # Only `$ALL_CAPS` env vars are supported
        assert config["password"] == yml_config["password"]

        # Values of extras can reference regular settings
        assert config["_extra"] == config["multiple"]
        assert config["_extra_generic"] == config["foo"]

        # Env vars inside env var values do not get expanded
        assert config["test"] == "$FOO"

        # Expansion can be disabled
        config = subject.as_dict(session=session, expand_env_vars=False)
        assert yml_config == {
            key: value for key, value in config.items() if key in yml_config
        }

    @pytest.mark.order(3)
    @pytest.mark.usefixtures("tap")
    def test_nested_keys(self, session, subject, project):
        def set_config(path, value):
            subject.set(path, value, store=SettingValueStore.MELTANO_YML)

        def unset_config(path):
            subject.unset(path, store=SettingValueStore.MELTANO_YML)

        def yml_config():
            with project.meltano_update() as meltano:
                extractor = meltano.plugins.extractors[0]
                return extractor.config

        def final_config():
            return subject.as_dict(session=session)

        set_config("metadata.stream.replication-key", "created_at")

        assert yml_config()["metadata.stream.replication-key"] == "created_at"
        assert final_config()["metadata.stream.replication-key"] == "created_at"

        set_config(["metadata", "stream", "replication-key"], "created_at")

        yml = yml_config()
        assert "metadata.stream.replication-key" not in yml
        assert yml["metadata"]["stream"]["replication-key"] == "created_at"
        assert final_config()["metadata.stream.replication-key"] == "created_at"

        set_config(["metadata", "stream", "replication-method"], "INCREMENTAL")

        yml = yml_config()
        assert "metadata.stream.replication-key" not in yml
        assert "metadata.stream.replication-method" not in yml
        assert yml["metadata"]["stream"]["replication-key"] == "created_at"
        assert yml["metadata"]["stream"]["replication-method"] == "INCREMENTAL"
        final = final_config()
        assert final["metadata.stream.replication-key"] == "created_at"
        assert final["metadata.stream.replication-method"] == "INCREMENTAL"

        set_config(["metadata.stream.replication-key"], "created_at")
        unset_config(["metadata.stream.replication-method"])

        yml = yml_config()
        assert "metadata" not in yml
        assert yml["metadata.stream.replication-key"] == "created_at"
        assert final_config()["metadata.stream.replication-key"] == "created_at"

        set_config(["metadata", "stream.replication-key"], "created_at")

        yml = yml_config()
        assert "metadata.stream.replication-key" not in yml
        assert yml["metadata"]["stream.replication-key"] == "created_at"
        assert final_config()["metadata.stream.replication-key"] == "created_at"

        unset_config(["metadata", "stream.replication-key"])

        yml = yml_config()
        assert "metadata.stream.replication-key" not in yml
        assert "metadata" not in yml
        assert "metadata.stream.replication-key" not in final_config()

    @pytest.mark.usefixtures("tap")
    def test_custom_setting(self, session, subject, env_var):
        subject.set("custom_string", "from_yml", store=SettingValueStore.MELTANO_YML)
        subject.set("custom_bool", True, store=SettingValueStore.MELTANO_YML)
        subject.set("custom_array", [1, 2, 3, "4"], store=SettingValueStore.MELTANO_YML)

        assert subject.get_with_source("custom_string", session=session) == (
            "from_yml",
            SettingValueStore.MELTANO_YML,
        )
        assert subject.get_with_source("custom_bool", session=session) == (
            True,
            SettingValueStore.MELTANO_YML,
        )
        assert subject.get_with_source("custom_array", session=session) == (
            [1, 2, 3, "4"],
            SettingValueStore.MELTANO_YML,
        )

        subject.env_override = {
            env_var(subject, "custom_string"): "from_env",
            env_var(subject, "custom_bool"): "off",
            env_var(subject, "custom_array"): '["foo"]',
        }

        assert subject.get_with_source("custom_string", session=session) == (
            "from_env",
            SettingValueStore.ENV,
        )
        assert subject.get_with_source("custom_bool", session=session) == (
            False,
            SettingValueStore.ENV,
        )
        assert subject.get_with_source("custom_array", session=session) == (
            ["foo"],
            SettingValueStore.ENV,
        )

    def test_date_values(self, subject, monkeypatch):
        today = datetime.now(timezone.utc).date()
        monkeypatch.setitem(subject.plugin.config, "start_date", today)
        assert subject.get("start_date") == today.isoformat()

        now = datetime.now(timezone.utc)
        monkeypatch.setitem(subject.plugin.config, "start_date", now)
        assert subject.get("start_date") == now.isoformat()

    @pytest.mark.usefixtures("tap")
    def test_kind_object(
        self,
        subject: PluginSettingsService,
        monkeypatch,
        env_var,
    ):
        assert subject.get_with_source("object") == (
            {"nested": "from_default"},
            SettingValueStore.DEFAULT,
        )

        subject.set("object.username", "from_meltano_yml")

        assert subject.get_with_source("object") == (
            {"username": "from_meltano_yml"},
            SettingValueStore.MELTANO_YML,
        )

        subject.set("data.password", "from_meltano_yml_alias")

        assert subject.get_with_source("object") == (
            {"username": "from_meltano_yml", "password": "from_meltano_yml_alias"},
            SettingValueStore.MELTANO_YML,
        )

        subject.set(["object", "password"], "from_meltano_yml")

        assert subject.get_with_source("object") == (
            {"username": "from_meltano_yml", "password": "from_meltano_yml"},
            SettingValueStore.MELTANO_YML,
        )

        subject.set(["object", "deep", "nesting"], "from_meltano_yml")

        assert subject.get_with_source("object") == (
            {
                "username": "from_meltano_yml",
                "password": "from_meltano_yml",
                "deep.nesting": "from_meltano_yml",
            },
            SettingValueStore.MELTANO_YML,
        )

        monkeypatch.setenv(env_var(subject, "object.deep.nesting"), "from_env")

        assert subject.get_with_source("object") == (
            {
                "username": "from_meltano_yml",
                "password": "from_meltano_yml",
                "deep.nesting": "from_env",
            },
            SettingValueStore.ENV,
        )

        assert subject.as_dict(process=False)["object"] == {
            "username": "from_meltano_yml",
            "password": "from_meltano_yml",
            "deep.nesting": "from_env",
        }
        assert subject.as_dict(process=True)["object"] == {
            "username": "from_meltano_yml",
            "password": "from_meltano_yml",
            "deep": {
                "nesting": "from_env",
            },
        }

        monkeypatch.setenv(env_var(subject, "data"), '{"foo":"from_env_alias"}')

        assert subject.get_with_source("object") == (
            {"foo": "from_env_alias"},
            SettingValueStore.ENV,
        )

        monkeypatch.setenv(env_var(subject, "object"), '{"foo":"from_env"}')

        assert subject.get_with_source("object") == (
            {"foo": "from_env"},
            SettingValueStore.ENV,
        )

    @pytest.mark.usefixtures("tap")
    def test_extra(self, subject, monkeypatch, env_var):
        subject._setting_defs = None

        assert "_select" in subject.as_dict()
        assert "_select" in subject.as_dict(extras=True)
        assert "_select" not in subject.as_dict(extras=False)

        assert subject.get_with_source("_select") == (
            ["*.*"],
            SettingValueStore.DEFAULT,
        )

        monkeypatch.setitem(
            subject.plugin.parent._variant.extras,
            "select",
            ["from_default"],
        )
        subject._setting_defs = None

        assert subject.get_with_source("_select") == (
            ["from_default"],
            SettingValueStore.DEFAULT,
        )

        monkeypatch.setitem(
            subject.plugin.config,
            "_select",
            ["from_meltano_yml_config"],
        )

        assert subject.get_with_source("_select") == (
            ["from_meltano_yml_config"],
            SettingValueStore.MELTANO_YML,
        )

        monkeypatch.setitem(subject.plugin.extras, "select", ["from_meltano_yml_extra"])

        assert subject.get_with_source("_select") == (
            ["from_meltano_yml_extra"],
            SettingValueStore.MELTANO_YML,
        )

        subject.set("_select", ["from_meltano_yml"])

        assert subject.get_with_source("_select") == (
            ["from_meltano_yml"],
            SettingValueStore.MELTANO_YML,
        )

        subject.unset("_select")

        assert subject.get_with_source("_select") == (
            ["from_default"],
            SettingValueStore.DEFAULT,
        )

        monkeypatch.setenv(env_var(subject, "_select"), '["from_env"]')

        assert subject.get_with_source("_select") == (
            ["from_env"],
            SettingValueStore.ENV,
        )

    @pytest.mark.usefixtures("environment")
    def test_extra_object(
        self,
        subject,
        monkeypatch,
        env_var,
        project_add_service,
        plugin_settings_service_factory,
    ):
        try:
            transform = project_add_service.add(
                PluginType.TRANSFORMS,
                "tap-mock-transform",
            )
        except PluginAlreadyAddedException as err:
            transform = err.plugin

        subject = plugin_settings_service_factory(transform)
        assert "_vars" in subject.as_dict()
        assert "_vars" in subject.as_dict(extras=True)
        assert "_vars" not in subject.as_dict(extras=False)

        assert subject.get_with_source("_vars") == ({}, SettingValueStore.DEFAULT)

        monkeypatch.setitem(
            subject.plugin.parent._variant.extras,
            "vars",
            {"var": "from_default", "other": "from_default"},
        )
        subject._setting_defs = None

        assert subject.get_with_source("_vars") == (
            {"var": "from_default", "other": "from_default"},
            SettingValueStore.DEFAULT,
        )

        monkeypatch.setitem(subject.plugin.extras, "vars", {"var": "from_meltano_yml"})

        assert subject.get_with_source("_vars") == (
            {"var": "from_meltano_yml", "other": "from_default"},
            SettingValueStore.MELTANO_YML,
        )

        subject.set(
            "_vars",
            {"other": "from_meltano_yml"},
            store=SettingValueStore.MELTANO_YML,
        )

        assert subject.get_with_source("_vars") == (
            {"var": "from_default", "other": "from_meltano_yml"},
            SettingValueStore.MELTANO_YML,
        )

        with monkeypatch.context() as patch:
            patch.setenv(env_var(subject, "_vars.var"), "from_env")
            assert subject.get_with_source("_vars") == (
                {"var": "from_env", "other": "from_meltano_yml"},
                SettingValueStore.ENV,
            )

        with monkeypatch.context() as patch:
            patch.setenv(env_var(subject, "_vars"), '{"var": "from_env"}')
            assert subject.get_with_source("_vars") == (
                {"var": "from_env"},
                SettingValueStore.ENV,
            )

        subject.set(
            "_vars",
            {"dev_setting": "from_dev_env"},
            store=SettingValueStore.MELTANO_ENV,
        )
        assert subject.get_with_source("_vars") == (
            {
                "var": "from_default",
                "other": "from_meltano_yml",
                "dev_setting": "from_dev_env",
            },
            SettingValueStore.MELTANO_ENV,
        )

        subject.project.deactivate_environment()

        inherited = project_add_service.add(
            PluginType.TRANSFORMS,
            "tap-mock-transform--inherited",
            inherit_from=transform.name,
        )
        inherited_subject = plugin_settings_service_factory(inherited)
        inherited_subject.set("_vars", {"new": "from_inheriting"})

        assert inherited_subject.get_with_source("_vars") == (
            {
                "other": "from_meltano_yml",
                "new": "from_inheriting",
            },
            SettingValueStore.MELTANO_YML,
        )

    def test_find_setting_raises_with_conflicting(
        self,
        tap,
        plugin_settings_service_factory,
        monkeypatch,
    ):
        subject = plugin_settings_service_factory(tap)
        monkeypatch.setenv("TAP_MOCK_ALIASED", "value_0")
        monkeypatch.setenv("TAP_MOCK_ALIASED_1", "value_1")
        with pytest.raises(ConflictingSettingValueException):
            subject.get("aliased")

    def test_find_setting_raises_with_multiple(
        self,
        tap,
        plugin_settings_service_factory,
        monkeypatch,
    ):
        subject = plugin_settings_service_factory(tap)
        monkeypatch.setenv("TAP_MOCK_ALIASED", "value_0")
        monkeypatch.setenv("TAP_MOCK_ALIASED_1", "value_0")
        with pytest.raises(MultipleEnvVarsSetException):
            subject.get("aliased")

    def test_find_setting_aliases(self, tap, plugin_settings_service_factory):
        subject = plugin_settings_service_factory(tap)
        subject.set("aliased_3", "value_3")
        assert subject.get("aliased") == "value_3"

    @pytest.mark.order(-1)
    def test_strict_env_var_mode_on_raises_error(self, subject):
        subject.project_settings_service.set(
            [FEATURE_FLAG_PREFIX, str(FeatureFlags.STRICT_ENV_VAR_MODE)],
            True,
        )
        subject.set("stacked_env_var", "${NONEXISTENT_ENV_VAR}")
        with pytest.raises(EnvironmentVariableNotSetError):
            subject.get("stacked_env_var")

    @pytest.mark.order(-1)
    def test_strict_env_var_mode_off_no_raise_error(self, subject):
        subject.project_settings_service.set(
            [FEATURE_FLAG_PREFIX, str(FeatureFlags.STRICT_ENV_VAR_MODE)],
            False,
        )
        subject.set("stacked_env_var", "${NONEXISTENT_ENV_VAR}")
        assert subject.get("stacked_env_var") is None



from __future__ import annotations

from pathlib import Path

from meltano.core.tracking import schemas

schemas_dir = (
    Path(__file__).parents[4]
    / "src"
    / "meltano"
    / "core"
    / "tracking"
    / "iglu-client-embedded"
    / "schemas"
    / "com.meltano"
)

versions_in_use = {
    x.name: x.version.split("-")
    for x in vars(schemas).values()  # noqa: WPS421
    if isinstance(x, schemas.IgluSchema)
}

versions_available = {
    schema_dir.stem: max(
        x.stem.split("-") for x in (schema_dir / "jsonschema").iterdir()
    )
    for schema_dir in schemas_dir.iterdir()
}


def test_using_latest_schemas():
    assert versions_in_use == versions_available



"""Defines helpers for use by the CLI."""

from __future__ import annotations

import logging
import os
import signal
import sys
import typing as t
from contextlib import contextmanager
from enum import Enum, auto

import click
import structlog
from click_default_group import DefaultGroup
from click_didyoumean import DYMGroup

from meltano.core.error import MeltanoConfigurationError
from meltano.core.logging import setup_logging
from meltano.core.plugin.base import PluginDefinition, PluginType
from meltano.core.plugin.error import InvalidPluginDefinitionError, PluginNotFoundError
from meltano.core.plugin_install_service import (
    PluginInstallReason,
    PluginInstallService,
    PluginInstallState,
    PluginInstallStatus,
)
from meltano.core.plugin_lock_service import LockfileAlreadyExistsError
from meltano.core.project_add_service import (
    PluginAddedReason,
    PluginAlreadyAddedException,
    ProjectAddService,
)
from meltano.core.setting_definition import SettingKind
from meltano.core.tracking.contexts import CliContext, CliEvent, ProjectContext

if sys.version_info < (3, 11):
    ReprEnum = Enum
else:
    from enum import ReprEnum

if t.TYPE_CHECKING:
    from meltano.core.plugin.base import PluginRef
    from meltano.core.plugin.project_plugin import ProjectPlugin
    from meltano.core.project import Project
    from meltano.core.project_plugins_service import ProjectPluginsService

setup_logging()

logger = structlog.stdlib.get_logger(__name__)


class CliError(Exception):
    """CLI Error."""

    def __init__(self, *args, **kwargs):
        """Instantiate custom CLI Error exception."""
        super().__init__(*args, **kwargs)

        self.printed = False

    def print(self):  # noqa: T202
        """Print CLI error."""
        if self.printed:
            return

        logger.debug(str(self), exc_info=True)
        click.secho(str(self), fg="red", err=True)

        self.printed = True


def print_added_plugin(
    plugin: ProjectPlugin,
    reason: PluginAddedReason = PluginAddedReason.USER_REQUEST,
    update: bool = False,
) -> None:
    """Print added plugin."""
    descriptor = plugin.type.descriptor
    if reason is PluginAddedReason.RELATED:
        descriptor = f"related {descriptor}"
    elif reason is PluginAddedReason.REQUIRED:
        descriptor = f"required {descriptor}"

    action, preposition = (
        (
            "Updated" if plugin.should_add_to_file() else "Updating",
            "in",
        )
        if update
        else (
            "Added" if plugin.should_add_to_file() else "Adding",
            "to",
        )
    )

    click.secho(
        f"{action} {descriptor} '{plugin.name}' {preposition} your project",
        fg="green",
    )

    inherit_from = plugin.inherit_from
    has_variant = plugin.is_variant_set
    variant_label = plugin.variant_label(plugin.variant)
    if inherit_from:
        if has_variant:
            inherit_from = f"{inherit_from}, variant {variant_label}"

        click.echo(f"Inherit from:\t{inherit_from}")
    elif has_variant:
        click.echo(f"Variant:\t{variant_label}")

    if repo_url := plugin.repo:
        click.echo(f"Repository:\t{repo_url}")

    if docs_url := plugin.docs:
        click.echo(f"Documentation:\t{docs_url}")


def _prompt_plugin_namespace(plugin_type, plugin_name):
    click.secho(
        f"Adding new custom {plugin_type.descriptor} with name '{plugin_name}'...",
        fg="green",
    )
    click.echo(
        f"\nSpecify the plugin's {click.style('namespace', fg='blue')}, which "
        "will serve as the:\n"
        "- identifier to find related/compatible plugins",
    )
    if plugin_type == PluginType.EXTRACTORS:
        click.echo(
            "- default database schema (`load_schema` extra),"
            "  for use by loaders that support a target schema",
        )
    elif plugin_type == PluginType.LOADERS:
        click.echo(
            "- default target database dialect (`dialect` extra),"
            "  for use by transformers that connect with the database",
        )
    click.echo(
        "\nHit Return to accept the default: plugin name with underscores "
        "instead of dashes\n",
    )

    return click.prompt(
        click.style("(namespace)", fg="blue"),
        type=str,
        default=plugin_name.replace("-", "_"),
    )


def _prompt_plugin_pip_url(plugin_name: str) -> str | None:
    click.echo(
        "\nSpecify the plugin's "
        f"{click.style('`pip install` argument', fg='blue')}, for example:"
        "- PyPI package name:\n"
        f"\t{plugin_name}\n"
        "- Git repository URL:\n"
        "\tgit+https://<PLUGIN REPO URL>.git\n"
        "- local directory, in editable/development mode:\n"
        f"\t-e extract/{plugin_name}\n"
        "- 'n' if using a local executable (nothing to install)\n\n"
        "Default: plugin name as PyPI package name\n",
    )
    result = click.prompt(
        click.style("(pip_url)", fg="blue"),
        type=str,
        default=plugin_name,
    )
    return None if result == "n" else result


def _prompt_plugin_executable(pip_url: str | None, plugin_name: str) -> str:
    derived_from = "`pip_url`"
    prompt_request = "executable name"
    if pip_url is None:
        derived_from = "the plugin name"
        prompt_request = "executable path"
    click.echo(
        f"\nSpecify the plugin's {click.style(prompt_request, fg='blue')}\n"
        f"\nDefault: name derived from {derived_from}\n",
    )
    plugin_basename = os.path.basename(pip_url or plugin_name)
    package_name, _ = os.path.splitext(plugin_basename)
    return click.prompt(click.style("(executable)", fg="blue"), default=package_name)


def _prompt_plugin_capabilities(plugin_type):
    if plugin_type != PluginType.EXTRACTORS:
        return []

    click.echo(
        f"\nSpecify the tap's {click.style('supported Singer features', fg='blue')} "
        "(executable flags), for example:\n"
        "\t`catalog`: supports the `--catalog` flag\n"
        "\t`discover`: supports the `--discover` flag\n"
        "\t`properties`: supports the `--properties` flag\n"
        "\t`state`: supports the `--state` flag\n\n"
        "To find out what features a tap supports, reference its "
        "documentation or try one of the tricks under"
        "https://docs.meltano.com/guide/integration#troubleshooting.\n\n"
        "Multiple capabilities can be separated using commas.\n\n"
        "Default: no capabilities\n",
    )

    return click.prompt(
        click.style("(capabilities)", fg="blue"),
        type=list,
        default=[],
        value_proc=lambda value: [word.strip() for word in value.split(",")]
        if value
        else [],
    )


def _prompt_plugin_settings(plugin_type: PluginType) -> list[dict[str, t.Any]]:
    if plugin_type not in {
        PluginType.EXTRACTORS,
        PluginType.LOADERS,
        PluginType.TRANSFORMERS,
    }:
        return []

    click.echo(
        f"\nSpecify the {plugin_type.descriptor}'s "
        f"{click.style('supported settings', fg='blue')}\n"
        "Multiple setting names (keys) can be separated using commas.\n\n"
        "A setting kind can be specified alongside the name (key) by using "
        "the `:` delimiter,\n"
        "e.g. `port:integer` to set the kind `integer` for the name `port`\n\n"
        "Supported setting kinds:",
    )
    click.echo(
        " | ".join([click.style(kind.value, fg="magenta") for kind in SettingKind]),
    )
    click.echo(
        "\n- Credentials and other sensitive setting types should use the "
        f"{click.style('password', fg='magenta')} kind.\n"
        "- If not specified, setting kind defaults to "
        f"{click.style('string', fg='magenta')}.\n"
        "- Nested properties can be represented using the `.` separator, "
        'e.g. `auth.username` for `{ "auth": { "username": value } }`.\n'
        f"- To find out what settings a {plugin_type.descriptor} supports, "
        "reference its documentation.\n"
        "\nDefault: no settings\n",
    )

    settings = None
    while settings is None:  # noqa:  WPS426  # allows lambda in loop
        settings_input = click.prompt(
            click.style("(settings)", fg="blue"),
            type=list,
            default=[],
            value_proc=lambda value: [
                setting.strip().partition(":") for setting in value.split(",")
            ]
            if value
            else [],
        )
        try:
            settings = [
                {"name": name, "kind": kind and SettingKind(kind).value}
                for name, sep, kind in settings_input
            ]
        except ValueError as ex:
            click.secho(str(ex), fg="red")

    return settings


def add_plugin(  # noqa: C901
    plugin_type: PluginType,
    plugin_name: str,
    *,
    python: str | None = None,
    add_service: ProjectAddService,
    variant=None,
    inherit_from=None,
    custom=False,
    update=False,
    lock=True,
    plugin_yaml=None,
):
    """Add Plugin to given Project."""
    if custom:
        # XXX: For backwards compatibility, the namespace must be prompted for first.
        namespace = _prompt_plugin_namespace(plugin_type, plugin_name)
        pip_url = _prompt_plugin_pip_url(plugin_name)
        plugin_attrs = {
            "namespace": namespace,
            "pip_url": pip_url,
            "executable": _prompt_plugin_executable(pip_url, plugin_name),
            "capabilities": _prompt_plugin_capabilities(plugin_type),
            "settings": _prompt_plugin_settings(plugin_type),
        }
    else:
        plugin_attrs = {}

    if plugin_yaml is not None:
        try:
            plugin_definition = PluginDefinition.parse(
                {
                    "plugin_type": plugin_type,
                    **plugin_yaml,
                },
            )
        except TypeError as e:
            raise InvalidPluginDefinitionError(plugin_yaml) from e

        # exclude unspecified properties
        plugin_definition.extras.clear()

        plugin_attrs = plugin_definition.canonical()  # type: ignore[assignment]

        plugin_name = plugin_attrs.pop("name")
        variant = plugin_attrs.pop("variant", variant)

    plugin: ProjectContext | PluginRef

    try:
        plugin = add_service.add(
            plugin_type,
            plugin_name,
            variant=variant,
            inherit_from=inherit_from,
            lock=lock,
            update=update,
            python=python,
            **plugin_attrs,
        )
        print_added_plugin(plugin, update=update)
    except PluginAlreadyAddedException as err:
        plugin = err.plugin
        click.secho(
            (
                f"{plugin_type.descriptor.capitalize()} '{plugin_name}' "
                "already exists in your Meltano project"
            ),
            fg="yellow",
            err=True,
        )
        if variant and variant != plugin.variant:
            new_plugin = err.new_plugin
            click.echo(
                f"\nTo switch from the current '{plugin.variant}' variant "
                f"to '{new_plugin.variant}':\n"
                "1. Update `variant` and `pip_url` in your `meltano.yml` "
                "project file:\n"
                f"\tname: {plugin.name}\n"
                f"\tvariant: {new_plugin.variant}\n"
                f"\tpip_url: {new_plugin.pip_url}\n"
                "2. Reinstall the plugin:\n"
                f"\tmeltano install {plugin_type.singular} {plugin.name}\n"
                "3. Check if the configuration is still valid (and make "
                "changes until it is):\n"
                f"\tmeltano config {plugin.name} list\n\n"
                "To learn more, visit "
                "https://docs.meltano.com/guide/plugin-management#switching-from-one-variant-to-another\n\n"  # noqa: E501
                f"Alternatively, to keep the existing '{plugin.name}' with "
                f"variant '{new_plugin.variant}', add variant '{new_plugin.variant}' "
                "as a separate plugin with its own unique name:\n"
                f"\tmeltano add {plugin_type.singular} "
                f"{plugin.name}--{new_plugin.variant} --inherit-from {plugin.name} "
                f"--variant {new_plugin.variant}\n\n"
                "To learn more, visit "
                "https://docs.meltano.com/guide/plugin-management#multiple-variants",
            )
        else:
            click.echo(
                "To add it to your project another time so that each can be "
                "configured differently,\n"
                "add a new plugin inheriting from the existing one with its "
                "own unique name:\n"
                f"\tmeltano add {plugin_type.singular} {plugin.name}--new "
                f"--inherit-from {plugin.name}",
            )
    except LockfileAlreadyExistsError as exc:
        # TODO: This is a BasePlugin, not a ProjectPlugin, as this method
        # should return! Results in `KeyError: venv_name`
        plugin = exc.plugin
        click.secho(
            f"Plugin definition is already locked at {exc.path}.",
            fg="yellow",
        )
        click.echo(
            "You can remove the file manually to avoid using a stale definition.",
        )

    click.echo()

    return plugin


def add_required_plugins(
    plugins: list[ProjectPlugin],
    add_service: ProjectAddService,
    lock: bool = True,
):
    """Add any Plugins required by the given Plugin."""
    added_plugins = []
    for plugin_install in plugins:
        required_plugins = add_service.add_required(
            plugin_install,
            lock=lock,
        )
        for required_plugin in required_plugins:
            print_added_plugin(required_plugin, reason=PluginAddedReason.REQUIRED)
            click.echo()

        added_plugins.extend(required_plugins)

    return added_plugins


def install_status_update(install_state: PluginInstallState):
    """
    Print the status of plugin installation.

    Used as the callback for PluginInstallService.
    """
    plugin = install_state.plugin
    desc = plugin.type.descriptor

    if install_state.status is PluginInstallStatus.SKIPPED:
        level = (
            logging.DEBUG
            if install_state.reason == PluginInstallReason.AUTO
            else logging.INFO
        )
        logger.log(level, "%s %s '%s'", install_state.verb, desc, plugin.name)
    elif install_state.status in {
        PluginInstallStatus.RUNNING,
        PluginInstallStatus.SUCCESS,
    }:
        logger.info("%s %s '%s'", install_state.verb, desc, plugin.name)
    elif install_state.status is PluginInstallStatus.ERROR:
        logger.error(install_state.message)
        logger.info(install_state.details)
    elif install_state.status is PluginInstallStatus.WARNING:  # pragma: no cover
        logger.warning(install_state.message)


async def install_plugins(
    project,
    plugins,
    *,
    reason=PluginInstallReason.INSTALL,
    parallelism=None,
    clean=False,
    force=False,
) -> bool:
    """Install the provided plugins and report results to the console."""
    install_service = PluginInstallService(
        project,
        status_cb=install_status_update,
        parallelism=parallelism,
        clean=clean,
        force=force,
    )
    install_results = await install_service.install_plugins(plugins, reason=reason)
    num_successful = len([status for status in install_results if status.successful])
    num_skipped = len([status for status in install_results if status.skipped])
    num_failed = len(install_results) - num_successful

    level = logging.INFO
    if num_failed >= 0 and num_successful == 0:
        level = logging.ERROR
    elif num_failed > 0 and num_successful > 0:
        level = logging.WARNING
    elif reason == PluginInstallReason.AUTO and num_skipped == len(plugins):
        level = logging.DEBUG

    if len(plugins) > 1:
        logger.log(
            level,
            "%s %d/%d plugins",
            "Updated" if reason == PluginInstallReason.UPGRADE else "Installed",
            num_successful - num_skipped,
            num_successful + num_failed,
        )
    if num_skipped:  # pragma: no cover
        logger.log(
            level,
            "Skipped installing %d/%d plugins",
            num_skipped,
            num_successful + num_failed,
        )

    return num_failed == 0


@contextmanager
def propagate_stop_signals(proc):
    """Propagate stop signals to `proc`, then wait for it to terminate."""

    def _handler(sig, _):  # noqa: WPS430
        proc.send_signal(sig)
        logger.debug("stopping child process...")
        # unset signal handler, so that even if the child never stops,
        # an additional stop signal will terminate as usual
        signal.signal(signal.SIGTERM, original_termination_handler)

    original_termination_handler = signal.signal(signal.SIGTERM, _handler)
    try:
        yield
    finally:
        signal.signal(signal.SIGTERM, original_termination_handler)


def check_dependencies_met(
    plugin_refs: list[PluginRef],
    plugins_service: ProjectPluginsService,
) -> tuple[bool, str]:
    """Check dependencies of added plugins are met.

    Args:
        plugins: List of plugins to be added.
        plugin_service: Plugin service to use when checking for dependencies.

    Returns:
        A tuple with dependency check outcome (True/False), and a string
        message with details of the check.
    """
    passed = True
    messages = []

    for plugin in plugin_refs:
        if plugin.type == PluginType.TRANSFORMS:
            # check that the `dbt` transformer plugin is installed
            try:
                plugins_service.get_transformer()
            except PluginNotFoundError:
                passed = False
                messages.append(
                    f"Plugin '{plugin.name}' requires a transformer plugin. "
                    "Please first add a transformer using `meltano add transformer`.",
                )
    if passed:
        message = "All dependencies met"
    else:
        message = f"Dependencies not met: {'; '.join(messages)}"
    return passed, message


def activate_environment(
    ctx: click.Context,
    project: Project,
    required: bool = False,
) -> None:
    """Activate the selected environment.

    The selected environment is whatever was selected with the `--environment`
    option, or the default environment (set in `meltano.yml`) otherwise.

    Args:
        ctx: The Click context, used to determine the selected environment.
        project: The project for which the environment will be activated.
    """
    if ctx.obj.get("selected_environment"):
        project.activate_environment(ctx.obj["selected_environment"])
        # Update the project context being used for telemetry:
        project_ctx = next(
            ctx
            for ctx in ctx.obj["tracker"].contexts
            if isinstance(ctx, ProjectContext)
        )
        project_ctx.environment_name = ctx.obj["selected_environment"]

    elif required:
        raise MeltanoConfigurationError(
            reason="A Meltano environment must be specified",
            instruction="Set the `default_environment` option in "
            "`meltano.yml`, or the `--environment` CLI option",
        )


def activate_explicitly_provided_environment(
    ctx: click.Context,
    project: Project,
) -> None:
    """Activate the selected environment if it has been explicitly set.

    Some commands (e.g. `config`, `job`, etc.) do not respect the configured
    `default_environment`, and will only run with an environment active if it
    has been explicitly set (e.g. with the `--environment` CLI option).

    Args:
        ctx: The Click context, used to determine the selected environment.
        project: The project for which the environment will be activated.
    """
    if ctx.obj.get("is_default_environment"):
        logger.info(
            f"The default environment {ctx.obj['selected_environment']!r} will "  # noqa: G004
            f"be ignored for `meltano {ctx.command.name}`. To configure a specific "
            "environment, please use the option `--environment=<environment name>`.",
        )
        project.deactivate_environment()
    else:
        activate_environment(ctx, project)


class CliEnvironmentBehavior(Enum):
    """Enum of the different Meltano environment activation behaviours."""

    # Use explicit environment, or `default_environment`, or fail.
    environment_required = auto()

    # Use explicit environment, or `default_environment`, or no environment.
    environment_optional_use_default = auto()

    # Use explicit environment, or no environment; ignore `default_environment`.
    environment_optional_ignore_default = auto()


def enact_environment_behavior(
    behavior: CliEnvironmentBehavior | None,
    ctx: click.Context,
) -> None:
    """Activate the environment in the specified way."""
    if behavior is None or "selected_environment" not in ctx.obj:
        return
    if behavior is CliEnvironmentBehavior.environment_optional_use_default:
        activate_environment(ctx, ctx.obj["project"], required=False)
    elif behavior is CliEnvironmentBehavior.environment_required:
        activate_environment(ctx, ctx.obj["project"], required=True)
    elif behavior is CliEnvironmentBehavior.environment_optional_ignore_default:
        activate_explicitly_provided_environment(ctx, ctx.obj["project"])


class InstrumentedCmdMixin:
    """Shared functionality for all instrumented commands."""

    def __init__(
        self,
        *args,
        environment_behavior: CliEnvironmentBehavior | None = None,
        **kwargs,
    ):
        """Initialize the `InstrumentedCmdMixin`.

        Args:
            args: Arguments to pass to the parent class.
            environment_behavior: The behavior to use regarding the activation
                of the Meltano environment for this command.
            kwargs: Keyword arguments to pass to the parent class.
        """
        self.environment_behavior = environment_behavior
        super().__init__(*args, **kwargs)


class InstrumentedGroupMixin(InstrumentedCmdMixin):
    """Shared functionality for all instrumented groups."""

    def invoke(self, ctx: click.Context):
        """Update the telemetry context and invoke the group."""
        ctx.ensure_object(dict)
        enact_environment_behavior(self.environment_behavior, ctx)
        if ctx.obj.get("tracker"):
            ctx.obj["tracker"].add_contexts(CliContext.from_click_context(ctx))
        # Typing these mixin hierarchies is a bit messy, so we'll just ignore it here
        # https://mypy.readthedocs.io/en/latest/more_types.html#mixin-classes
        super().invoke(ctx)  # type: ignore[misc]


class InstrumentedDefaultGroup(InstrumentedGroupMixin, DefaultGroup, DYMGroup):
    """Click group with telemetry instrumentation and a default command."""


class InstrumentedGroup(InstrumentedGroupMixin, DYMGroup):
    """Click group with telemetry instrumentation."""


class InstrumentedCmd(InstrumentedCmdMixin, click.Command):
    """Click command that automatically fires telemetry events when invoked.

    Both starting and ending events are fired. The ending event fired is
    dependent on whether invocation of the command resulted in an Exception.
    """

    def invoke(self, ctx: click.Context):
        """Invoke the requested command firing start and events accordingly."""
        ctx.ensure_object(dict)
        enact_environment_behavior(self.environment_behavior, ctx)
        if ctx.obj.get("tracker"):
            tracker = ctx.obj["tracker"]
            tracker.add_contexts(CliContext.from_click_context(ctx))
            tracker.track_command_event(CliEvent.started)
            try:
                super().invoke(ctx)
            except Exception:
                tracker.track_command_event(CliEvent.failed)
                raise
            tracker.track_command_event(CliEvent.completed)
        else:
            super().invoke(ctx)


class PartialInstrumentedCmd(InstrumentedCmdMixin, click.Command):
    """Click command with partial telemetry instrumentation.

    Only automatically fires a 'start' event.
    """

    def invoke(self, ctx):
        """Invoke the requested command firing only a start event."""
        ctx.ensure_object(dict)
        enact_environment_behavior(self.environment_behavior, ctx)
        if ctx.obj.get("tracker"):
            ctx.obj["tracker"].add_contexts(CliContext.from_click_context(ctx))
            ctx.obj["tracker"].track_command_event(CliEvent.started)
        super().invoke(ctx)


# TODO: Use StrEnum (and its backport) when we can drop Python 3.8 support
class AutoInstallBehavior(str, ReprEnum):
    """Enum of the different behaviors for automatic plugin installation."""

    install = "install"
    no_install = "no_install"
    only_install = "only_install"



"""Test the lock CLI command."""

from __future__ import annotations

import typing as t

import pytest

from meltano.cli import cli
from meltano.cli.utils import CliError
from meltano.core.plugin_lock_service import PluginLock

if t.TYPE_CHECKING:
    from click.testing import CliRunner

    from meltano.core.hub import MeltanoHubService
    from meltano.core.plugin.project_plugin import ProjectPlugin
    from meltano.core.project import Project


class TestLock:
    @pytest.mark.order(0)
    @pytest.mark.usefixtures("project")
    @pytest.mark.parametrize(
        "args",
        (
            ["lock"],
            ["lock", "--all", "tap-mock"],
        ),
        ids=("noop", "all-and-plugin-name"),
    )
    def test_lock_invalid_options(self, cli_runner: CliRunner, args: list[str]):
        result = cli_runner.invoke(cli, args)
        assert result.exit_code == 1

        exception_message = "Exactly one of --all or plugin name must be specified."
        assert exception_message == str(result.exception)

    @pytest.mark.order(1)
    @pytest.mark.usefixtures("tap", "target")
    def test_lockfile_exists(
        self,
        cli_runner: CliRunner,
        project: Project,
    ):
        lockfiles = list(project.root_plugins_dir().glob("./*/*.lock"))
        assert len(lockfiles) == 2

        result = cli_runner.invoke(cli, ["lock", "--all"])
        assert result.exit_code == 0
        assert "Lockfile exists for extractor tap-mock" in result.stdout
        assert "Lockfile exists for loader target-mock" in result.stdout
        assert "Locked definition" not in result.stdout

    @pytest.mark.order(2)
    def test_lockfile_update(
        self,
        cli_runner: CliRunner,
        project: Project,
        tap: ProjectPlugin,
        hub_endpoints: MeltanoHubService,
    ):
        tap_lock = PluginLock(project, tap)

        assert tap_lock.path.exists()
        old_checksum = tap_lock.sha256_checksum
        old_definition = tap_lock.load()

        # Update the plugin in Hub
        hub_endpoints["/extractors/tap-mock--meltano"]["settings"].append(
            {
                "name": "foo",
                "value": "bar",
            },
        )

        result = cli_runner.invoke(cli, ["lock", "--all", "--update"])
        assert result.exit_code == 0
        assert result.stdout.count("Lockfile exists") == 0
        assert result.stdout.count("Locked definition") == 2

        new_checksum = tap_lock.sha256_checksum
        new_definition = tap_lock.load()
        assert new_checksum != old_checksum
        assert len(new_definition.settings) == len(old_definition.settings) + 1

        new_setting = new_definition.settings[-1]
        assert new_setting.name == "foo"
        assert new_setting.value == "bar"

    @pytest.mark.order(3)
    @pytest.mark.usefixtures("tap", "inherited_tap", "hub_endpoints")
    def test_lockfile_update_extractors(
        self,
        cli_runner: CliRunner,
        project: Project,
    ):
        lockfiles = list(project.root_plugins_dir().glob("./*/*.lock"))
        # 1 tap, 1 target
        assert len(lockfiles) == 2

        result = cli_runner.invoke(
            cli,
            ["lock", "--all", "--update", "--plugin-type", "extractor"],
        )
        assert result.exit_code == 0
        assert "Lockfile exists" not in result.stdout
        assert "Locked definition for extractor tap-mock" in result.stdout
        assert "Extractor tap-mock-inherited is an inherited plugin" in result.stdout

    def test_lock_plugin_not_found(self, cli_runner: CliRunner):
        result = cli_runner.invoke(cli, ["lock", "not-a-plugin"])
        assert result.exit_code == 1
        assert isinstance(result.exception, CliError)
        assert "No matching plugin(s) found" in str(result.exception)



"""Utilities for parsing string lists of plugins.

Turns string lists of plugins into `BlockSet` and `PluginCommand` instances.
"""

from __future__ import annotations

import typing as t

import click

from meltano.core.block.blockset import BlockSet, BlockSetValidationError
from meltano.core.block.extract_load import ELBContextBuilder, ExtractLoadBlocks
from meltano.core.block.plugin_command import PluginCommandBlock, plugin_command_invoker
from meltano.core.block.singer import CONSUMERS, SingerBlock
from meltano.core.plugin import PluginType
from meltano.core.plugin.error import PluginNotFoundError
from meltano.core.task_sets_service import TaskSetsService

if t.TYPE_CHECKING:
    import uuid

    import structlog

    from meltano.core.plugin.project_plugin import ProjectPlugin


def is_command_block(plugin: ProjectPlugin) -> bool:
    """Check if a plugin is a command block.

    Args:
        plugin: Plugin to check.

    Returns:
        True if plugin is a command block.
    """
    return plugin.type not in {
        PluginType.EXTRACTORS,
        PluginType.LOADERS,
        PluginType.MAPPERS,
    }


def validate_block_sets(
    log: structlog.BoundLogger,
    blocks: list[BlockSet | PluginCommandBlock],
) -> bool:
    """Perform validation of all blocks in a list that implement the BlockSet interface.

    Args:
        log: Logger to use in the event of a validation error.
        blocks: A list of blocks.

    Returns:
        True if all blocks are valid, False otherwise.
    """
    for idx, blk in enumerate(blocks):
        if blk == BlockSet:
            log.debug("validating ExtractLoadBlock.", set_number=idx)
            try:
                blk.validate_set()
            except Exception as err:
                log.error("Validation failed.", err=err)
                return False
    return True


class BlockParser:  # noqa: D101
    def __init__(
        self,
        log: structlog.BoundLogger,
        project,
        blocks: list[str],
        full_refresh: bool | None = False,
        refresh_catalog: bool | None = False,
        no_state_update: bool | None = False,
        force: bool | None = False,
        state_id_suffix: str | None = None,
        merge_state: bool | None = False,
        run_id: uuid.UUID | None = None,
    ):
        """
        Parse a meltano run command invocation into a list of blocks.

        Args:
            log: Logger to use.
            project: Project to use.
            blocks: List of block names to parse.
            full_refresh: Whether to perform a full refresh (applies to all
                found sets).
            refresh_catalog: Whether to ignore cached catalog.
            no_state_update: Whether to run with or without state updates.
            force: Whether to force a run if a job is already running (applies
                to all found sets).
            state_id_suffix: State ID suffix to use.
            merge_state: Whether to merge state at end of run.
            run_id: Custom run ID to use.

        Raises:
            ClickException: If a block name is not found.
        """
        self.log = log
        self.project = project

        self._full_refresh = full_refresh
        self._refresh_catalog = refresh_catalog
        self._no_state_update = no_state_update
        self._force = force
        self._state_id_suffix = state_id_suffix
        self._plugins: list[ProjectPlugin] = []
        self._commands: dict[int, str] = {}
        self._mappings_ref: dict[int, str] = {}
        self._merge_state = merge_state
        self._run_id = run_id

        task_sets_service: TaskSetsService = TaskSetsService(project)

        blocks = self._expand_jobs(blocks, task_sets_service)

        for idx, name in enumerate(blocks):
            try:
                parsed_name, command_name = name.split(":")
            except ValueError:
                parsed_name = name
                command_name = None

            try:
                plugin = self.project.plugins.find_plugin(parsed_name)
            except PluginNotFoundError as e:
                raise click.ClickException(f"Block {name} not found") from e  # noqa: EM102

            if plugin and task_sets_service.exists(name):
                raise click.ClickException(
                    f"Ambiguous reference to '{name}' which matches a job "  # noqa: EM102
                    "name AND a plugin name.",
                )

            if plugin.type == PluginType.MAPPERS:
                self._mappings_ref[idx] = parsed_name

            self._plugins.append(plugin)
            if command_name:
                self._commands[idx] = command_name
                self.log.debug(
                    "plugin command added for execution",
                    commands=self._commands,
                    command_name=command_name,
                    plugin_name=parsed_name,
                )

            self.log.debug("found plugin in cli invocation", plugin_name=plugin.name)

    @property
    def plugins(self) -> list[ProjectPlugin]:
        """Return the list of plugins in the block.

        Returns:
            A list of ProjectPlugin.
        """
        return self._plugins

    def _expand_jobs(self, blocks: list[str], task_sets: TaskSetsService) -> list[str]:
        """Expand any jobs present in a list of blocks into their raw block names.

        Example:
            Given a job named "somejob" which consists of a single task of "tap target":
            ["somejob", "dbt:run"] -> ["tap", "target", "dbt:run"]

        Args:
            blocks: List of block names to parse.
            task_sets: TaskSetsService to use.

        Returns:
            List of block names with jobs expanded.
        """
        expanded_blocks: list[str] = []
        for name in blocks:
            if task_sets.exists(name):
                self.log.debug(
                    "expanding job to tasks",
                    job_name=name,
                    tasks=task_sets.get(name).flat_args,
                )
                expanded_blocks.extend(task_sets.get(name).flat_args)
            else:
                expanded_blocks.append(name)
        return expanded_blocks

    def find_blocks(
        self,
        offset: int = 0,
    ) -> t.Generator[BlockSet | PluginCommandBlock | ExtractLoadBlocks, None, None]:
        """Find all blocks in the invocation.

        Args:
            offset: Offset to start from.

        Yields:
            Blocks (either BlockSet or PluginCommandBlock).

        Raises:
            BlockSetValidationError: If unknown command is found or if a
                unexpected block sequence is found.
        """
        cur = offset
        while cur < len(self._plugins):
            plugin = self._plugins[cur]
            elb, idx = self._find_next_elb_set(cur)
            if elb:
                self.log.debug("found ExtractLoadBlocks set", offset=cur)
                yield elb
                cur += idx
            elif is_command_block(plugin):
                self.log.debug(
                    "found PluginCommand",
                    offset=cur,
                    plugin_type=plugin.type,
                )
                yield plugin_command_invoker(
                    self._plugins[cur],
                    self.project,
                    command=self._commands.get(cur),
                )
                cur += 1
            else:
                raise BlockSetValidationError(
                    "Unknown command type or bad block sequence at index "  # noqa: EM102
                    f"{cur + 1}, starting block '{plugin.name}'",  # noqa: WPS237
                )

    def _find_next_elb_set(  # noqa: WPS231, WPS213
        self,
        offset: int = 0,
    ) -> tuple[ExtractLoadBlocks | None, int]:  # noqa: WPS231, WPS213
        """Search plugins to find an extract EL block set.

        Args:
            offset: Optional starting offset for search.

        Returns:
            The `ExtractLoadBlocks` object, and offset for remaining plugins.

        Raises:
            BlockSetValidationError: If the block set is not valid.
        """
        blocks: list[SingerBlock] = []

        builder = (
            ELBContextBuilder(self.project)
            .with_force(self._force)
            .with_full_refresh(self._full_refresh)
            .with_refresh_catalog(self._refresh_catalog)
            .with_no_state_update(self._no_state_update)
            .with_state_id_suffix(self._state_id_suffix)
            .with_merge_state(self._merge_state)
            .with_run_id(self._run_id)
        )

        if self._plugins[offset].type != PluginType.EXTRACTORS:
            self.log.debug(
                "next block not extractor",
                offset=offset,
                plugin_type=self._plugins[offset].type,
            )
            return None, offset

        self.log.debug(
            "head of set is extractor as expected",
            block=self._plugins[offset].name,
        )

        blocks.append(builder.make_block(self._plugins[offset]))

        for idx, plugin in enumerate(self._plugins[offset + 1 :]):  # noqa: E203
            next_block = idx + 1

            if plugin.type not in CONSUMERS:
                self.log.debug(
                    "next block not a consumer of output",
                    offset=offset,
                    plugin_type=plugin.type,
                )
                return None, offset + next_block

            self.log.debug("found block", block_type=plugin.type, index=next_block)

            if plugin.type == PluginType.MAPPERS:
                self.log.debug(
                    "found mapper",
                    plugin_type=plugin.type,
                    plugin_name=plugin.name,
                    mapping=self._mappings_ref.get(next_block),
                    idx=next_block,
                )
                # Checks to see if the mapper plugin name is the same as the
                # mappings name. If they both match then a validation error is
                # raised because the meltano run command needs the mappings
                # name to obtain the settings to pass to the parent mapper
                # plugin. We also want to fail if the user names them the same
                # to stop errors due to ambiguous commands.
                if plugin.name == self._mappings_ref.get(next_block):
                    self.log.warning(
                        "Found unexpected mapper plugin name. ",
                        plugin_name=plugin.name,
                    )
                    raise BlockSetValidationError(
                        f"Expected unique mappings name not the mapper plugin "  # noqa: EM102
                        f"name: {plugin.name}.",
                    )
                blocks.append(builder.make_block(plugin))

            elif plugin.type == PluginType.LOADERS:
                self.log.debug("blocks", offset=offset, idx=next_block)
                blocks.append(builder.make_block(plugin))
                elb = ExtractLoadBlocks(builder.context(), blocks)
                return elb, idx + 2
            else:
                self.log.warning(
                    "Found unexpected plugin type for block in middle of block set.",
                    plugin_type=plugin.type,
                    plugin_name=plugin.name,
                )
                raise BlockSetValidationError(
                    f"Expected {PluginType.MAPPERS} or {PluginType.LOADERS}.",  # noqa: EM102
                )
        raise BlockSetValidationError("Found no end in block set!")  # noqa: EM101



from __future__ import annotations

import json

import pytest

from asserts import assert_cli_runner
from meltano.cli import cli
from meltano.core.plugin.singer.catalog import (
    ListSelectedExecutor,
    SelectedNode,
    SelectionType,
)
from meltano.core.select_service import SelectService


class TestCliSelect:
    @pytest.mark.usefixtures("project")
    def test_update_select_pattern(self, cli_runner, tap):
        # add select pattern
        result = cli_runner.invoke(
            cli,
            ["--no-environment", "select", tap.name, "mock", "*"],
        )
        assert_cli_runner(result)
        # verify pattern was added
        result = cli_runner.invoke(cli, ["config", "--extras", tap.name])
        assert_cli_runner(result)
        json_config = json.loads(result.stdout)
        assert "mock.*" in json_config["_select"]
        # remove select pattern
        result = cli_runner.invoke(
            cli,
            ["--no-environment", "select", tap.name, "--rm", "mock", "*"],
        )
        assert_cli_runner(result)
        # verify select pattern removed
        result = cli_runner.invoke(cli, ["config", "--extras", tap.name])
        assert_cli_runner(result)
        json_config = json.loads(result.stdout)
        assert "mock.*" not in json_config["_select"]

    @pytest.mark.usefixtures("project")
    def test_select_list(self, cli_runner, tap, monkeypatch: pytest.MonkeyPatch):
        async def mock_list_all(*args, **kwargs):  # noqa: ARG001
            result = ListSelectedExecutor()
            result.streams = {
                SelectedNode(key="users", selection=SelectionType.SELECTED)
            }
            result.properties = {
                "users": {
                    SelectedNode(key="id", selection=SelectionType.SELECTED),
                    SelectedNode(key="name", selection=SelectionType.EXCLUDED),
                }
            }
            return result

        monkeypatch.setattr(
            SelectService,
            "list_all",
            mock_list_all,
        )

        # list selection
        result = cli_runner.invoke(
            cli,
            [
                "--no-environment",
                "select",
                tap.name,
                "--list",
                "--all",
            ],
        )
        assert_cli_runner(result)

        assert "[selected ] users.id" in result.stdout
        assert "[excluded ] users.name" in result.stdout



from __future__ import annotations

import json
import logging
import os
import typing as t
from collections import Counter, abc
from copy import deepcopy
from http import HTTPStatus

import pytest
import requests
from requests.adapters import BaseAdapter

from meltano.core.plugin.base import PluginType

if t.TYPE_CHECKING:
    from meltano.core.project import Project

logging.basicConfig(level=logging.INFO)

PYTEST_BACKEND = os.getenv("PYTEST_BACKEND", "sqlite")

pytest_plugins = [
    "fixtures.db",
    "fixtures.fs",
    "fixtures.core",
    "fixtures.cli",
    "fixtures.docker",
]

if PYTEST_BACKEND == "sqlite":
    pytest_plugins.append("fixtures.db.sqlite")
elif PYTEST_BACKEND == "postgresql":
    pytest_plugins.append("fixtures.db.postgresql")
elif PYTEST_BACKEND == "postgresql_psycopg3":
    pytest_plugins.append("fixtures.db.postgresql_psycopg3")
elif PYTEST_BACKEND == "mssql":
    pytest_plugins.append("fixtures.db.mssql")
else:
    raise Exception(f"Unsuported backend: {PYTEST_BACKEND}.")  # noqa: EM102

BACKEND = ["sqlite", "postgresql", "mssql", "mysql"]


def pytest_runtest_setup(item):
    backend_marker = item.get_closest_marker("backend")

    # currently, there is no distinction between the SYSTEM database
    # and the WAREHOUSE database at the test level.
    # There is only one database used for the tests, and it serves
    # both as SYSTEM and WAREHOUSE.
    if backend_marker and backend_marker.args[0] != PYTEST_BACKEND:
        pytest.skip()


@pytest.fixture(scope="session")
def concurrency():
    return {
        "threads": int(os.getenv("PYTEST_CONCURRENCY_THREADS", 8)),
        "processes": int(os.getenv("PYTEST_CONCURRENCY_PROCESSES", 8)),
        "cases": int(os.getenv("PYTEST_CONCURRENCY_CASES", 64)),  # noqa: WPS432
    }


class MockAdapter(BaseAdapter):
    RETURN_500 = {
        "/extractors/this-returns-500--original": {"error": "Server error"},
    }

    def _process_discovery(self, api_url: str, discovery: dict) -> dict:
        hub = {}
        for plugin_type in PluginType:
            index_key = f"/{plugin_type}/index"
            hub[index_key] = {}
            for plugin in discovery.get(plugin_type, []):
                plugin_name = plugin["name"]
                hub[index_key][plugin_name] = {"variants": {}}
                default_variant = None

                variants = plugin.pop("variants", [])

                for variant in variants:
                    variant_name = variant["name"]

                    if not default_variant or variant.get("default", False):
                        default_variant = variant_name

                    hub[index_key][plugin_name]["variants"][variant_name] = {
                        "ref": (
                            f"{api_url}/plugins/{plugin_type}/"
                            f"{plugin_name}--{variant_name}"
                        ),
                    }

                    plugin_key = f"/{plugin_type}/{plugin_name}--{variant_name}"
                    hub[plugin_key] = {
                        **plugin,
                        **variant,
                        "name": plugin_name,
                        "label": plugin.get("label"),
                        "namespace": plugin["namespace"],
                        "variant": variant_name,
                    }

                if not variants:
                    variant_name = plugin.get("variant", "original")
                    default_variant = variant_name

                    hub[index_key][plugin_name]["variants"][variant_name] = {
                        "ref": (
                            f"{api_url}/plugins/{plugin_type}/"
                            f"{plugin_name}--{variant_name}"
                        ),
                    }

                    plugin_key = f"/{plugin_type}/{plugin_name}--{variant_name}"
                    hub[plugin_key] = {
                        **plugin,
                        "name": plugin_name,
                        "namespace": plugin["namespace"],
                        "variant": variant_name,
                        "settings": plugin.get("settings", []),
                        "commands": plugin.get("commands", {}),
                        "docs": plugin.get("docs"),
                        "repo": plugin.get("repo"),
                        "pip_url": plugin.get("pip_url"),
                    }

                    if "dialect" in plugin:
                        hub[plugin_key]["dialect"] = plugin["dialect"]

                hub[index_key][plugin_name]["logo_url"] = (
                    f"https://mock.meltano.com/{plugin_name}.png"
                )
                hub[index_key][plugin_name]["default_variant"] = default_variant

        return hub

    def __init__(self, api_url: str, discovery: dict) -> None:
        """Create a mock HTTP adapter for the Hub.

        Args:
            api_url: The base URL of the Hub.
            discovery: A parsed discovery.yml file.
        """
        super().__init__()
        self.api_url = api_url
        self.count = Counter()
        self._mapping = self._process_discovery(api_url, deepcopy(discovery))

        # Special cases
        self._mapping["/extractors/index"]["this-returns-500"] = {
            "default_variant": "original",
            "logo_url": "https://mock.meltano.com/this-returns-500.png",
            "variants": {
                "original": {
                    "ref": f"{api_url}/plugins/extractors/this-returns-500--original",
                },
            },
        }

    def send(
        self,
        request: requests.PreparedRequest,
        stream: bool = False,  # noqa: ARG002
        timeout: float  # noqa: ARG002, WPS320
        | tuple[float, float]
        | tuple[float, None]
        | None = None,
        verify: bool | str = True,  # noqa: ARG002
        cert: t.Any | None = None,  # noqa: ARG002
        proxies: abc.Mapping[str, str] | None = None,  # noqa: ARG002
    ):
        _, endpoint = request.path_url.split("/meltano/api/v1/plugins")

        response = requests.Response()
        response.request = request
        response.url = request.url

        response_500 = self.RETURN_500.get(endpoint)
        if response_500:
            response.status_code = HTTPStatus.INTERNAL_SERVER_ERROR
            response.reason = "Internal Server Error"
            response._content = json.dumps(response_500).encode()
            return response

        try:
            data = self._mapping[endpoint]
        except KeyError:
            response.status_code = HTTPStatus.NOT_FOUND
            response._content = b""
            return response

        self.count[endpoint] += 1
        response.status_code = HTTPStatus.OK
        response._content = json.dumps(data).encode()
        return response


@pytest.fixture(scope="class")
def hub_mock_adapter(discovery) -> t.Callable[[str], MockAdapter]:
    def _get_adapter(api_url: str) -> MockAdapter:
        return MockAdapter(api_url, discovery)

    return _get_adapter


@pytest.fixture(scope="class")
def hub_endpoints(project: Project):
    adapter = project.hub_service.session.adapters[project.hub_service.hub_api_url]
    return adapter._mapping


@pytest.fixture()
def hub_request_counter(project: Project):
    counter: Counter = project.hub_service.session.get_adapter(
        project.hub_service.hub_api_url,
    ).count
    counter.clear()
    return counter



from __future__ import annotations

import datetime
import json
import typing as t
import uuid

from sqlalchemy.dialects.postgresql import UUID
from sqlalchemy.ext.mutable import MutableDict
from sqlalchemy.orm import mapped_column
from sqlalchemy.types import CHAR, INTEGER, VARCHAR, DateTime, TypeDecorator
from typing_extensions import Annotated


class JSONEncodedDict(TypeDecorator):
    """Represents an immutable structure as a json-encoded string.

    Usage:
        JSONEncodedDict(VARCHAR_LENGTH)
    """

    impl = VARCHAR
    cache_ok = True

    def process_bind_param(  # noqa: D102
        self,
        value,
        dialect,  # noqa: ARG002
    ):
        if value is not None:
            value = json.dumps(value)

        return value

    def process_result_value(  # noqa: D102
        self,
        value,
        dialect,  # noqa: ARG002
    ):
        if value is not None:
            value = json.loads(value)
        return value


class IntFlag(TypeDecorator):
    impl = INTEGER
    cache_ok = True

    # force the cast to INTEGER
    def process_bind_param(  # noqa: D102
        self,
        value,
        dialect,  # noqa: ARG002
    ):
        return int(value)


class GUID(TypeDecorator):
    """Platform-independent GUID type.

    Uses PostgreSQL's UUID type, otherwise uses
    CHAR(32), storing as stringified hex values.

    Reference:
    https://docs.sqlalchemy.org/en/13/core/custom_types.html#backend-agnostic-guid-type
    """

    impl = CHAR
    cache_ok = True

    def load_dialect_impl(self, dialect):
        if dialect.name == "postgresql":
            return dialect.type_descriptor(UUID())
        type_descriptor_length = 32
        return dialect.type_descriptor(CHAR(type_descriptor_length))

    def process_bind_param(self, value, dialect):
        if value is None:
            return value
        if dialect.name == "postgresql":
            return str(value)
        if not isinstance(value, uuid.UUID):
            value = uuid.UUID(value)
        return value.hex

    def process_result_value(  # noqa: D102
        self,
        value,
        dialect,  # noqa: ARG002
    ):
        if value is None:
            return value
        if not isinstance(value, uuid.UUID):
            value = uuid.UUID(value)
        return value


class DateTimeUTC(TypeDecorator):
    """Parses datetimes timezone-aware and stores them as UTC."""

    impl = DateTime
    cache_ok = True

    def process_bind_param(self, value: datetime.datetime | None, _dialect: str):
        """Convert the datetime value to UTC and remove the timezone.

        Args:
            value: The datetime value to convert.

        Returns:
            The converted datetime value.
        """
        if value is None:
            return None

        if value.tzinfo:
            value = value.astimezone(datetime.timezone.utc)
        return value.replace(tzinfo=None)

    def process_result_value(self, value: datetime.datetime | None, _dialect: str):
        """Convert the naive datetime value to UTC.

        Args:
            value: The datetime value to convert.

        Returns:
            The converted datetime value.
        """
        if value is not None:
            value = value.replace(tzinfo=datetime.timezone.utc)
        return value


GUIDType = Annotated[uuid.UUID, mapped_column(GUID, default=uuid.uuid4)]
StateType = Annotated[
    t.Dict[str, str],
    mapped_column(MutableDict.as_mutable(JSONEncodedDict)),
]
IntPK = Annotated[int, mapped_column(primary_key=True)]
StrPK = Annotated[str, mapped_column(primary_key=True)]



"""CLI command `meltano install`."""

from __future__ import annotations

import typing as t

import click
import structlog

from meltano.cli.params import pass_project
from meltano.cli.utils import CliError, PartialInstrumentedCmd, install_plugins
from meltano.core.block.parser import BlockParser
from meltano.core.plugin import PluginType
from meltano.core.schedule_service import ScheduleService
from meltano.core.tracking.contexts import CliEvent, PluginsTrackingContext
from meltano.core.utils import run_async

if t.TYPE_CHECKING:
    from meltano.core.project import Project
    from meltano.core.tracking import Tracker

ANY = "-"

logger = structlog.getLogger(__name__)


@click.command(cls=PartialInstrumentedCmd, short_help="Install project dependencies.")
@click.argument(
    "plugin_type",
    type=click.Choice((*PluginType.cli_arguments(), ANY)),
    required=False,
)
@click.argument("plugin_name", nargs=-1, required=False)
@click.option(
    "--clean",
    is_flag=True,
    help="Completely reinstall a plugin rather than simply upgrading if necessary.",
)
@click.option(
    "--parallelism",
    "-p",
    type=click.INT,
    default=None,
    help=(
        "Limit the number of plugins to install in parallel. "
        "Defaults to the number of cores."
    ),
)
@click.option(
    "--force",
    "-f",
    is_flag=True,
    help="Ignore the required Python version declared by the plugins.",
)
@click.option(
    "--schedule",
    "-s",
    "schedule_name",
    help="Install all plugins from the given schedule.",
)
@click.pass_context
@pass_project(migrate=True)
@run_async
async def install(  # noqa: C901
    project: Project,
    ctx: click.Context,
    plugin_type: str,
    plugin_name: str,
    clean: bool,
    parallelism: int,
    force: bool,
    schedule_name: str,
):
    """
    Install all the dependencies of your project based on the meltano.yml file.

    \b\nRead more at https://docs.meltano.com/reference/command-line-interface#install
    """
    tracker: Tracker = ctx.obj["tracker"]
    try:
        if plugin_type and plugin_type != ANY:
            plugin_type = PluginType.from_cli_argument(plugin_type)
            plugins = project.plugins.get_plugins_of_type(plugin_type)
        else:
            plugins = list(project.plugins.plugins())

        if plugin_name:
            plugins = [plugin for plugin in plugins if plugin.name in plugin_name]

        if schedule_name:
            schedule_plugins = _get_schedule_plugins(
                ctx.obj["project"],
                schedule_name,
            )
            plugins = list(set(plugins) & set(schedule_plugins))
    except Exception:
        tracker.track_command_event(CliEvent.aborted)
        raise

    logger.info("Installing %d plugins", len(plugins))
    tracker.add_contexts(
        PluginsTrackingContext([(candidate, None) for candidate in plugins]),
    )
    tracker.track_command_event(CliEvent.inflight)

    success = await install_plugins(
        project,
        plugins,
        parallelism=parallelism,
        clean=clean,
        force=force,
    )
    if not success:
        tracker.track_command_event(CliEvent.failed)
        raise CliError("Failed to install plugin(s)")  # noqa: EM101
    tracker.track_command_event(CliEvent.completed)


def _get_schedule_plugins(project: Project, schedule_name: str):
    schedule_service = ScheduleService(project)
    schedule_obj = schedule_service.find_schedule(schedule_name)
    schedule_plugins = set()
    if schedule_obj.elt_schedule:
        for plugin_name in (schedule_obj.extractor, schedule_obj.loader):
            schedule_plugins.add(project.plugins.find_plugin(plugin_name))
    else:
        task_sets = schedule_service.task_sets_service.get(schedule_obj.job)
        for blocks in task_sets.flat_args_per_set:
            parser = BlockParser(logger, project, blocks)
            for plugin in parser.plugins:
                schedule_plugins.add(
                    project.plugins.find_plugin(plugin.info.get("name"))
                    if plugin.type == PluginType.MAPPERS
                    else plugin,
                )
    return schedule_plugins



from __future__ import annotations

import asyncio
import json
import re
import typing as t
import uuid

import click
import pytest
from mock import AsyncMock, mock

from meltano.cli import cli
from meltano.cli.run import UUIDParamType
from meltano.core.block.ioblock import IOBlock
from meltano.core.logging.utils import default_config
from meltano.core.plugin import PluginType
from meltano.core.plugin.singer import SingerTap
from meltano.core.plugin_invoker import PluginInvoker
from meltano.core.project_plugins_service import (
    AmbiguousMappingName,
    PluginAlreadyAddedException,
)

if t.TYPE_CHECKING:
    from meltano.core.logging.job_logging_service import JobLoggingService
    from meltano.core.project import Project


class MockIOBlock(IOBlock):
    string_id = "mock-io-block"


@pytest.fixture(scope="class")
def tap_mock_transform(project_add_service):
    try:
        return project_add_service.add(PluginType.TRANSFORMS, "tap-mock-transform")
    except PluginAlreadyAddedException as err:
        return err.plugin


@pytest.fixture()
def process_mock_factory():
    def _factory(name):
        process_mock = mock.Mock()
        process_mock.name = name
        process_mock.wait = AsyncMock(return_value=0)
        process_mock.returncode = 0
        process_mock.stdin.wait_closed = AsyncMock(return_value=True)
        return process_mock

    return _factory


@pytest.fixture()
def tap_process(process_mock_factory, tap):
    tap = process_mock_factory(tap)
    tap.stdout.at_eof.side_effect = (False, False, False, True)
    tap.stdout.readline = AsyncMock(side_effect=(b"SCHEMA\n", b"RECORD\n", b"STATE\n"))
    tap.stderr.at_eof.side_effect = (False, False, False, True)
    tap.stderr.readline = AsyncMock(
        side_effect=(b"tap starting\n", b"tap running\n", b"tap done\n"),
    )
    return tap


@pytest.fixture()
def target_process(process_mock_factory, target):
    target = process_mock_factory(target)

    # Have `target.wait` take 2s to make sure the tap always finishes before the target
    async def wait_mock():
        await asyncio.sleep(2)
        return target.wait.return_value

    target.wait.side_effect = wait_mock

    target.stdout.at_eof.side_effect = (False, False, False, True)
    target.stdout.readline = AsyncMock(
        side_effect=(b'{"line": 1}\n', b'{"line": 2}\n', b'{"line": 3}\n'),
    )
    target.stderr.at_eof.side_effect = (False, False, False, True)
    target.stderr.readline = AsyncMock(
        side_effect=(b"target starting\n", b"target running\n", b"target done\n"),
    )
    return target


@pytest.fixture()
def mapper_process(process_mock_factory, mapper):
    mapper = process_mock_factory(mapper)

    # Have `mapper.wait` take 1s to make sure the mapper always finishes after
    # the tap but before the target
    async def wait_mock():
        await asyncio.sleep(1)
        return mapper.wait.return_value

    mapper.wait.side_effect = wait_mock

    mapper.stdout.at_eof.side_effect = (False, False, False, True)
    mapper.stdout.readline = AsyncMock(
        side_effect=(b"SCHEMA\n", b"RECORD\n", b"STATE\n"),
    )
    mapper.stderr.at_eof.side_effect = (False, False, False, True)
    mapper.stderr.readline = AsyncMock(
        side_effect=(b"mapper starting\n", b"mapper running\n", b"mapper done\n"),
    )
    return mapper


@pytest.fixture()
def dbt_process(process_mock_factory, dbt):
    dbt = process_mock_factory(dbt)

    async def wait_mock():
        await asyncio.sleep(1)
        return dbt.wait.return_value

    dbt.wait.side_effect = wait_mock

    dbt.stdout.at_eof.side_effect = (False, True)
    dbt.stdout.readline = AsyncMock(side_effect=(b"Testoutput"))
    dbt.stderr.at_eof.side_effect = (False, False, False, True)
    dbt.stderr.readline = AsyncMock(
        side_effect=(b"dbt starting\n", b"dbt running\n", b"dbt done\n"),
    )
    return dbt


class EventMatcher:
    def __init__(self, result_output: str):
        """Build a matcher for the result output of a command."""
        self.seen_events: list[dict] = []
        self.seen_raw: list[str] = []

        for line in result_output.splitlines():
            try:
                parsed_line = json.loads(line)
            except json.JSONDecodeError:
                self.seen_raw.append(line)
                continue
            self.seen_events.append(parsed_line)

    def event_matches(self, event: str) -> bool:
        """Search result output for an event, that matches the given event.

        Args:
            event: the event to search for.

        Returns:
            True if the event was found, False otherwise.
        """
        for line in self.seen_events:  # noqa: RET503
            matches = line["event"] == event
            if matches:
                return True

    def find_by_event(self, event: str) -> list[dict] | None:
        """Return the first matching event, that matches the given event.

        Args:
            event: the event to search for.

        Returns:
            A list of matching events.
        """
        matches = []
        for line in self.seen_events:
            match = line["event"] == event
            if match:
                matches.append(line)
        return matches


class TestCliRunScratchpadOne:
    @pytest.mark.backend("sqlite")
    @pytest.mark.usefixtures("use_test_log_config", "project", "job_logging_service")
    def test_run_parsing_failures(
        self,
        cli_runner,
        tap,
        target,
        tap_process,
        target_process,
    ):
        result = cli_runner.invoke(cli, ["run"])
        assert result.exit_code == 0

        assert EventMatcher(result.stderr).event_matches("No valid blocks found.")

        args = ["run", tap.name]

        # exit cleanly when everything is fine
        create_subprocess_exec = AsyncMock(side_effect=(tap_process, target_process))

        # check that the various ELB validation checks actually run and fail as expected
        with mock.patch.object(SingerTap, "discover_catalog"), mock.patch.object(
            SingerTap,
            "apply_catalog_rules",
        ), mock.patch("meltano.core.plugin_invoker.asyncio") as asyncio_mock:
            asyncio_mock.create_subprocess_exec = create_subprocess_exec
            with pytest.raises(Exception, match="Found no end in block set!"):
                cli_runner.invoke(cli, args, catch_exceptions=False)

        args = ["run", tap.name, tap.name, target.name]
        with mock.patch.object(SingerTap, "discover_catalog"), mock.patch.object(
            SingerTap,
            "apply_catalog_rules",
        ), mock.patch("meltano.core.plugin_invoker.asyncio") as asyncio_mock2:
            asyncio_mock2.create_subprocess_exec = create_subprocess_exec
            with pytest.raises(
                Exception,
                match=(
                    "Unknown command type or bad block sequence at index 1, "
                    "starting block 'tap-mock'"
                ),
            ):
                cli_runner.invoke(cli, args, catch_exceptions=False)

        args = ["run", tap.name, target.name, target.name]
        with mock.patch.object(SingerTap, "discover_catalog"), mock.patch.object(
            SingerTap,
            "apply_catalog_rules",
        ), mock.patch("meltano.core.plugin_invoker.asyncio") as asyncio_mock3:
            asyncio_mock3.create_subprocess_exec = create_subprocess_exec
            with pytest.raises(
                Exception,
                match=(
                    "Unknown command type or bad block sequence at index 3, "
                    "starting block 'target-mock'"
                ),
            ):
                cli_runner.invoke(cli, args, catch_exceptions=False)

        # Verify that a vanilla ELB run works
        args = ["run", tap.name, target.name]
        with mock.patch.object(SingerTap, "discover_catalog"), mock.patch.object(
            SingerTap,
            "apply_catalog_rules",
        ), mock.patch("meltano.core.plugin_invoker.asyncio") as asyncio_mock4:
            asyncio_mock4.create_subprocess_exec = create_subprocess_exec
            result = cli_runner.invoke(cli, args, catch_exceptions=False)
            assert result.exit_code == 0

            matcher = EventMatcher(result.stderr)

            assert matcher.event_matches(
                "All ExtractLoadBlocks validated, starting execution.",
            )
            assert matcher.find_by_event("Block run completed.")[0]["success"]

    @pytest.mark.backend("sqlite")
    @pytest.mark.usefixtures(
        "use_test_log_config",
        "project",
        "mapper",
        "dbt",
        "job_logging_service",
    )
    def test_run_basic_invocations(
        self,
        cli_runner,
        tap,
        target,
        tap_process,
        target_process,
        mapper_process,
        dbt_process,
    ):
        # exit cleanly when everything is fine
        create_subprocess_exec = AsyncMock(
            side_effect=(tap_process, mapper_process, target_process),
        )

        # Verify that a vanilla ELB run works
        args = ["run", tap.name, "mock-mapping-0", target.name]
        with mock.patch.object(SingerTap, "discover_catalog"), mock.patch.object(
            SingerTap,
            "apply_catalog_rules",
        ), mock.patch("meltano.core.plugin_invoker.asyncio") as asyncio_mock:
            asyncio_mock.create_subprocess_exec = create_subprocess_exec
            result = cli_runner.invoke(cli, args, catch_exceptions=True)
            assert result.exit_code == 0

            matcher = EventMatcher(result.stderr)

            assert matcher.event_matches(
                "All ExtractLoadBlocks validated, starting execution.",
            )
            target_stop_event = matcher.find_by_event("target done")
            assert len(target_stop_event) == 1
            assert target_stop_event[0]["name"] == target.name
            assert target_stop_event[0]["cmd_type"] == "elb"
            assert target_stop_event[0]["stdio"] == "stderr"
            assert matcher.find_by_event("Block run completed.")[0]["success"]

        # Verify that a vanilla command plugin (dbt:run) run works
        invoke_async = AsyncMock(side_effect=(dbt_process,))  # dbt run
        args = ["run", "dbt:run"]
        with mock.patch.object(PluginInvoker, "invoke_async", new=invoke_async):
            result = cli_runner.invoke(cli, args, catch_exceptions=False)
            assert result.exit_code == 0

            matcher = EventMatcher(result.stderr)
            assert (
                matcher.find_by_event("found plugin in cli invocation")[0].get(
                    "plugin_name",
                )
                == "dbt"
            )
            dbt_start_event = matcher.find_by_event("dbt done")
            assert len(dbt_start_event) == 1
            assert dbt_start_event[0]["name"] == "dbt"
            assert dbt_start_event[0]["cmd_type"] == "command"
            assert dbt_start_event[0]["stdio"] == "stderr"
            assert matcher.find_by_event("Block run completed.")[0]["success"]

    @pytest.mark.backend("sqlite")
    @pytest.mark.usefixtures("use_test_log_config", "project")
    def test_run_custom_suffix_command_option(
        self,
        cli_runner,
        tap,
        target,
        tap_process,
        target_process,
        job_logging_service: JobLoggingService,
    ):
        # exit cleanly when everything is fine
        create_subprocess_exec = AsyncMock(side_effect=(tap_process, target_process))

        # Verify that a state ID with custom suffix from command option is
        # generated for an ELB run
        args = ["run", tap.name, target.name, "--state-id-suffix", "test-suffix"]

        with mock.patch.object(SingerTap, "discover_catalog"), mock.patch.object(
            SingerTap,
            "apply_catalog_rules",
        ), mock.patch("meltano.core.plugin_invoker.asyncio") as asyncio_mock:
            asyncio_mock.create_subprocess_exec = create_subprocess_exec
            result = cli_runner.invoke(cli, args, catch_exceptions=True)
            assert result.exit_code == 0

            matcher = EventMatcher(result.stderr)
            assert matcher.find_by_event("Block run completed.")[0]["success"]

            job_logging_service.get_latest_log(
                f"dev:{tap.name}-to-{target.name}:test-suffix",
            )

    @pytest.mark.backend("sqlite")
    @pytest.mark.usefixtures("use_test_log_config")
    @pytest.mark.parametrize(
        "suffix_args",
        (
            (
                "test-suffix",
                "test-suffix",
                [],
            ),
            (
                "${TEST_SUFFIX}",
                "test-suffix-single-env",
                [
                    ("TEST_SUFFIX", "test-suffix-single-env"),
                ],
            ),
            (
                "test-suffix-${TEST_SUFFIX_0}-${TEST_SUFFIX_1}",
                "test-suffix-multiple-env",
                [
                    ("TEST_SUFFIX_0", "multiple"),
                    ("TEST_SUFFIX_1", "env"),
                ],
            ),
        ),
        ids=[
            "static",
            "dynamic (single env)",
            "dynamic (multiple env)",
        ],
    )
    def test_run_custom_suffix_active_environment(
        self,
        suffix_args,
        cli_runner,
        project: Project,
        tap,
        target,
        tap_process,
        target_process,
        job_logging_service: JobLoggingService,
    ):
        state_id_suffix, expected_suffix, suffix_env = suffix_args

        # exit cleanly when everything is fine
        create_subprocess_exec = AsyncMock(side_effect=(tap_process, target_process))

        # Verify that a state ID with custom suffix from active environment is
        # generated for an ELB run
        project.activate_environment("dev")
        project.environment.state_id_suffix = state_id_suffix

        args = ["run", tap.name, target.name]

        with mock.patch.object(SingerTap, "discover_catalog"), mock.patch.object(
            SingerTap,
            "apply_catalog_rules",
        ), mock.patch(
            "meltano.core.plugin_invoker.asyncio",
        ) as asyncio_mock, pytest.MonkeyPatch().context() as mp:
            asyncio_mock.create_subprocess_exec = create_subprocess_exec

            for env in suffix_env:
                mp.setenv(*env)

            result = cli_runner.invoke(cli, args, catch_exceptions=True)
            assert result.exit_code == 0

            matcher = EventMatcher(result.stderr)
            assert matcher.find_by_event("Block run completed.")[0]["success"]

            job_logging_service.get_latest_log(
                f"dev:{tap.name}-to-{target.name}:${expected_suffix}",
            )

    @pytest.mark.backend("sqlite")
    @pytest.mark.usefixtures(
        "use_test_log_config",
        "project",
        "dbt",
        "job_logging_service",
    )
    def test_run_multiple_commands(self, cli_runner, dbt_process):
        # Verify that requesting the same command plugin multiple time with
        # different args works
        invoke_async = AsyncMock(
            side_effect=(
                dbt_process,
                dbt_process,
            ),
        )
        args = ["run", "dbt:test", "dbt:run"]
        with mock.patch.object(PluginInvoker, "invoke_async", new=invoke_async):
            result = cli_runner.invoke(cli, args, catch_exceptions=False)
            assert result.exit_code == 0

            matcher = EventMatcher(result.stderr)
            assert (
                matcher.find_by_event("found plugin in cli invocation")[0].get(
                    "plugin_name",
                )
                == "dbt"
            )

            command_add_events = matcher.find_by_event(
                "plugin command added for execution",
            )

            assert len(command_add_events) == 2
            assert invoke_async.call_count == 2

            assert command_add_events[0]["command_name"] == "test"
            assert invoke_async.mock_calls[0][2]["command"] == "test"

            assert command_add_events[1]["command_name"] == "run"
            assert invoke_async.mock_calls[1][2]["command"] == "run"

            assert matcher.find_by_event("Block run completed.")[0]["success"]
            assert matcher.find_by_event("Block run completed.")[1]["success"]

    @pytest.mark.backend("sqlite")
    @pytest.mark.usefixtures(
        "use_test_log_config",
        "project",
        "mapper",
        "dbt",
        "job_logging_service",
    )
    def test_run_complex_invocations(
        self,
        cli_runner,
        tap,
        target,
        tap_process,
        target_process,
        mapper_process,
        dbt_process,
    ):
        invoke_async = AsyncMock(
            side_effect=(tap_process, mapper_process, target_process, dbt_process),
        )
        args = ["run", tap.name, "mock-mapping-0", target.name, "dbt:run"]
        with mock.patch.object(PluginInvoker, "invoke_async", new=invoke_async):
            result = cli_runner.invoke(cli, args, catch_exceptions=False)
            assert result.exit_code == 0

            matcher = EventMatcher(result.stderr)
            assert matcher.event_matches(
                "found ExtractLoadBlocks set",
            )  # tap/target pair

            # make sure mapper was found and at its expected positions
            for ev in matcher.find_by_event("found block"):
                if ev["block_type"] == "mappers":
                    assert ev["index"] == 1

            assert (
                matcher.find_by_event("found PluginCommand")[0]["plugin_type"]
                == "transformers"
            )  # dbt

            completed_events = matcher.find_by_event("Block run completed.")
            assert len(completed_events) == 2
            for event in completed_events:
                assert event["success"]

            tap_stop_event = matcher.find_by_event("tap done")
            assert len(tap_stop_event) == 1
            assert tap_stop_event[0]["name"] == tap.name
            assert tap_stop_event[0]["cmd_type"] == "elb"
            assert tap_stop_event[0]["stdio"] == "stderr"

            target_stop_event = matcher.find_by_event("target done")
            assert len(target_stop_event) == 1
            assert target_stop_event[0]["name"] == target.name
            assert target_stop_event[0]["cmd_type"] == "elb"
            assert target_stop_event[0]["stdio"] == "stderr"

            dbt_done_event = matcher.find_by_event("dbt done")
            assert len(dbt_done_event) == 1
            assert dbt_done_event[0]["name"] == "dbt"
            assert dbt_done_event[0]["cmd_type"] == "command"
            assert dbt_done_event[0]["stdio"] == "stderr"

    @pytest.mark.backend("sqlite")
    @pytest.mark.usefixtures(
        "use_test_log_config",
        "project",
        "dbt",
        "job_logging_service",
    )
    def test_run_plugin_command_failure(
        self,
        cli_runner,
        tap,
        target,
        tap_process,
        target_process,
        dbt_process,
    ):
        args = ["run", tap.name, target.name, "dbt:run"]

        dbt_process.wait.return_value = 1
        dbt_process.returncode = 1
        dbt_process.stderr.readline.side_effect = (
            b"dbt starting\n",
            b"dbt running\n",
            b"dbt failure\n",
        )

        invoke_async = AsyncMock(side_effect=(tap_process, target_process, dbt_process))

        with mock.patch.object(PluginInvoker, "invoke_async", new=invoke_async):
            result = cli_runner.invoke(cli, args)
            assert result.exit_code == 1
            assert "`dbt run` failed" in str(result.exception)

            matcher = EventMatcher(result.stderr)
            assert matcher.event_matches(
                "found ExtractLoadBlocks set",
            )  # tap/target pair
            assert (
                matcher.find_by_event("found PluginCommand")[0]["plugin_type"]
                == "transformers"
            )  # dbt

            completed_events = matcher.find_by_event("Block run completed.")
            assert len(completed_events) == 2
            for event in completed_events:
                if event["block_type"] == "ExtractLoadBlocks":
                    assert event["success"]
                elif event["block_type"] == "InvokerCommand":
                    assert event["success"] is False

            tap_stop_event = matcher.find_by_event("tap done")
            assert len(tap_stop_event) == 1
            assert tap_stop_event[0]["name"] == tap.name
            assert tap_stop_event[0]["cmd_type"] == "elb"
            assert tap_stop_event[0]["stdio"] == "stderr"

            target_stop_event = matcher.find_by_event("target done")
            assert len(target_stop_event) == 1
            assert target_stop_event[0]["name"] == target.name
            assert target_stop_event[0]["cmd_type"] == "elb"
            assert target_stop_event[0]["stdio"] == "stderr"

            assert not matcher.event_matches("dbt done")
            assert matcher.event_matches("dbt starting")
            assert matcher.event_matches("dbt running")
            assert matcher.event_matches("dbt failure")

    @pytest.mark.backend("sqlite")
    @pytest.mark.usefixtures(
        "use_test_log_config",
        "project",
        "dbt",
        "job_logging_service",
    )
    def test_run_elb_tap_failure(
        self,
        cli_runner,
        tap,
        target,
        tap_process,
        target_process,
        dbt_process,
    ):
        # In this scenario, the tap fails on the third read. Target should still
        # complete, but dbt should not.
        args = ["run", tap.name, target.name, "dbt:run"]

        tap_process.wait.return_value = 1
        tap_process.returncode = 1
        tap_process.stderr.readline.side_effect = (
            b"tap starting\n",
            b"tap running\n",
            b"tap failure\n",
        )

        invoke_async = AsyncMock(side_effect=(tap_process, target_process, dbt_process))

        with mock.patch.object(PluginInvoker, "invoke_async", new=invoke_async):
            result = cli_runner.invoke(cli, args)

            assert (
                "Run invocation could not be completed as block failed: "
                "Extractor failed"
            ) in str(result.exception)
            assert result.exit_code == 1

            matcher = EventMatcher(result.stderr)
            assert matcher.event_matches("found ExtractLoadBlocks set")
            assert (
                matcher.find_by_event("found PluginCommand")[0]["plugin_type"]
                == "transformers"
            )

            completed_events = matcher.find_by_event("Block run completed.")
            assert len(completed_events) == 1
            assert completed_events[0]["success"] is False

            assert completed_events[0]["err"] == "RunnerError('Extractor failed')"
            assert completed_events[0]["exit_codes"]["extractors"] == 1

            tap_stop_event = matcher.find_by_event("tap failure")
            assert len(tap_stop_event) == 1
            assert tap_stop_event[0]["name"] == tap.name
            assert tap_stop_event[0]["cmd_type"] == "elb"
            assert tap_stop_event[0]["stdio"] == "stderr"

            target_stop_event = matcher.find_by_event("target done")
            assert len(target_stop_event) == 1
            assert target_stop_event[0]["name"] == target.name
            assert target_stop_event[0]["cmd_type"] == "elb"
            assert target_stop_event[0]["stdio"] == "stderr"

            # dbt should not have run at all
            assert not matcher.event_matches("dbt starting")
            assert not matcher.event_matches("dbt running")
            assert not matcher.event_matches("dbt done")

    @pytest.mark.backend("sqlite")
    @pytest.mark.usefixtures(
        "use_test_log_config",
        "project",
        "dbt",
        "job_logging_service",
    )
    def test_run_elb_target_failure_before_tap_finished(  # noqa: WPS118
        self,
        cli_runner,
        tap,
        target,
        tap_process,
        target_process,
        dbt_process,
    ):
        args = ["run", tap.name, target.name, "dbt:run"]

        # Have `tap_process.wait` take 2s to make sure the target can fail
        # before tap finishes
        async def tap_wait_mock():
            await asyncio.sleep(2)
            return tap_process.wait.return_value

        tap_process.wait.side_effect = tap_wait_mock

        # Writing to target stdin will fail because (we'll pretend) it has already died
        target_process.stdin = mock.Mock(spec=asyncio.StreamWriter)
        # capture_subprocess_output writer will return and close the pipe when
        # either BrokenPipeError or ConnectionResetError is enccountered it
        # does not itself reraise the exception - so you shouldn't expect to see these.
        target_process.stdin.write.side_effect = BrokenPipeError
        target_process.stdin.drain = AsyncMock(side_effect=ConnectionResetError)
        target_process.stdin.wait_closed = AsyncMock(return_value=True)

        # Have `target_process.wait` take 1s to make sure the
        # `stdin.write`/`drain` exceptions can be raised
        async def target_wait_mock():
            await asyncio.sleep(1)
            return 1

        target_process.wait.side_effect = target_wait_mock

        target_process.wait.return_value = 1
        target_process.returncode = 1
        target_process.stderr.readline.side_effect = (
            b"target starting\n",
            b"target running\n",
            b"target failure\n",
        )

        invoke_async = AsyncMock(side_effect=(tap_process, target_process, dbt_process))

        with mock.patch.object(PluginInvoker, "invoke_async", new=invoke_async):
            result = cli_runner.invoke(cli, args)

            assert (
                "Run invocation could not be completed as block failed: Loader failed"
                in str(result.exception)
            )
            assert result.exit_code == 1

            matcher = EventMatcher(result.stderr)
            assert matcher.event_matches(
                "found ExtractLoadBlocks set",
            )  # tap/target pair
            assert (
                matcher.find_by_event("found PluginCommand")[0]["plugin_type"]
                == "transformers"
            )  # dbt

            completed_events = matcher.find_by_event("Block run completed.")
            # there should only be one completed event
            assert len(completed_events) == 1
            assert completed_events[0]["success"] is False

            assert completed_events[0]["err"] == "RunnerError('Loader failed')"
            assert completed_events[0]["exit_codes"]["loaders"] == 1

            # The tap should NOT have finished, we'll have a write of the
            # SCHEMA message and then nothing further:
            assert matcher.event_matches("SCHEMA")
            assert not matcher.event_matches("RECORD")
            assert not matcher.event_matches("STATE")

            target_stop_event = matcher.find_by_event("target failure")
            assert len(target_stop_event) == 1
            assert target_stop_event[0]["name"] == target.name
            assert target_stop_event[0]["cmd_type"] == "elb"
            assert target_stop_event[0]["stdio"] == "stderr"

            # dbt should not have run at all
            assert not matcher.event_matches("dbt starting")
            assert not matcher.event_matches("dbt running")
            assert not matcher.event_matches("dbt done")

    @pytest.mark.backend("sqlite")
    @pytest.mark.usefixtures(
        "use_test_log_config",
        "project",
        "dbt",
        "job_logging_service",
    )
    def test_run_elb_target_failure_after_tap_finished(  # noqa: WPS118
        self,
        cli_runner,
        tap,
        target,
        tap_process,
        target_process,
        dbt_process,
    ):
        args = ["run", tap.name, target.name, "dbt:run"]

        target_process.wait.return_value = 1
        target_process.returncode = 1
        target_process.stderr.readline.side_effect = (
            b"target starting\n",
            b"target running\n",
            b"target failure\n",
        )

        invoke_async = AsyncMock(side_effect=(tap_process, target_process, dbt_process))

        with mock.patch.object(PluginInvoker, "invoke_async", new=invoke_async):
            result = cli_runner.invoke(cli, args)

            assert (
                "Run invocation could not be completed as block failed: Loader failed"
                in str(result.exception)
            )
            assert result.exit_code == 1

            matcher = EventMatcher(result.stderr)
            assert matcher.event_matches(
                "found ExtractLoadBlocks set",
            )  # tap/target pair
            assert (
                matcher.find_by_event("found PluginCommand")[0]["plugin_type"]
                == "transformers"
            )  # dbt

            completed_events = matcher.find_by_event("Block run completed.")
            # there should only be one completed event
            assert len(completed_events) == 1
            assert completed_events[0]["success"] is False
            assert completed_events[0]["err"] == "RunnerError('Loader failed')"
            assert completed_events[0]["exit_codes"]["loaders"] == 1

            tap_stop_event = matcher.find_by_event("tap done")
            assert len(tap_stop_event) == 1
            assert tap_stop_event[0]["name"] == tap.name
            assert tap_stop_event[0]["cmd_type"] == "elb"
            assert tap_stop_event[0]["stdio"] == "stderr"

            target_stop_event = matcher.find_by_event("target failure")
            assert len(target_stop_event) == 1
            assert target_stop_event[0]["name"] == target.name
            assert target_stop_event[0]["cmd_type"] == "elb"
            assert target_stop_event[0]["stdio"] == "stderr"

            # dbt should not have run at all
            assert not matcher.event_matches("dbt starting")
            assert not matcher.event_matches("dbt running")
            assert not matcher.event_matches("dbt done")

    @pytest.mark.backend("sqlite")
    @pytest.mark.usefixtures(
        "use_test_log_config",
        "project",
        "dbt",
        "job_logging_service",
    )
    def test_run_elb_tap_and_target_failed(
        self,
        cli_runner,
        tap,
        target,
        tap_process,
        target_process,
        dbt_process,
    ):
        args = ["run", tap.name, target.name, "dbt:run"]

        tap_process.wait.return_value = 1
        tap_process.returncode = 1
        tap_process.stderr.readline.side_effect = (
            b"tap starting\n",
            b"tap running\n",
            b"tap failure\n",
        )

        target_process.wait.return_value = 1
        target_process.returncode = 1
        target_process.stderr.readline.side_effect = (
            b"target starting\n",
            b"target running\n",
            b"target failure\n",
        )

        invoke_async = AsyncMock(side_effect=(tap_process, target_process, dbt_process))

        with mock.patch.object(PluginInvoker, "invoke_async", new=invoke_async):
            result = cli_runner.invoke(cli, args)

            assert (
                "Run invocation could not be completed as block failed: "
                "Extractor and loader failed"
            ) in str(result.exception)
            assert result.exit_code == 1

            matcher = EventMatcher(result.stderr)
            assert matcher.event_matches("found ExtractLoadBlocks set")
            assert (
                matcher.find_by_event("found PluginCommand")[0]["plugin_type"]
                == "transformers"
            )

            completed_events = matcher.find_by_event("Block run completed.")
            assert len(completed_events) == 1
            assert completed_events[0]["success"] is False

            assert (
                completed_events[0]["err"]
                == "RunnerError('Extractor and loader failed')"
            )
            assert completed_events[0]["exit_codes"]["loaders"] == 1

            tap_stop_event = matcher.find_by_event("tap failure")
            assert len(tap_stop_event) == 1
            assert tap_stop_event[0]["name"] == tap.name
            assert tap_stop_event[0]["cmd_type"] == "elb"
            assert tap_stop_event[0]["stdio"] == "stderr"

            target_stop_event = matcher.find_by_event("target failure")
            assert len(target_stop_event) == 1
            assert target_stop_event[0]["name"] == target.name
            assert target_stop_event[0]["cmd_type"] == "elb"
            assert target_stop_event[0]["stdio"] == "stderr"

            # dbt should not have run at all
            assert not matcher.event_matches("dbt starting")
            assert not matcher.event_matches("dbt running")
            assert not matcher.event_matches("dbt done")

    @pytest.mark.backend("sqlite")
    @pytest.mark.usefixtures(
        "use_test_log_config",
        "project",
        "dbt",
        "dbt_process",
        "job_logging_service",
    )
    def test_run_elb_tap_line_length_limit_error(
        self,
        cli_runner,
        tap,
        target,
        tap_process,
        target_process,
    ):
        args = ["run", tap.name, target.name]

        # Raise a `ValueError` wrapping a `LimitOverrunError`, like
        # `StreamReader.readline` does:
        # https://github.com/python/cpython/blob/v3.8.7/Lib/asyncio/streams.py#L549
        try:  # noqa: WPS328
            raise asyncio.LimitOverrunError(
                "Separator is not found, and chunk exceed the limit",  # noqa: EM101
                0,
            )
        except asyncio.LimitOverrunError as err:
            try:  # noqa: WPS328, WPS505
                # `ValueError` needs to be raised from inside the except block
                # for `LimitOverrunError` so that `__context__` is set.
                raise ValueError(str(err))
            except ValueError as wrapper_err:
                tap_process.stdout.readline.side_effect = wrapper_err

        # Have `tap_process.wait` take 1s to make sure the `LimitOverrunError`
        # exception can be raised before tap finishes
        async def wait_mock():
            await asyncio.sleep(1)
            return tap_process.wait.return_value

        tap_process.wait.side_effect = wait_mock

        invoke_async = AsyncMock(side_effect=(tap_process, target_process))
        with mock.patch.object(PluginInvoker, "invoke_async", new=invoke_async):
            result = cli_runner.invoke(cli, args)

            assert (
                "Run invocation could not be completed as block failed: "
                "Output line length limit exceeded"
            ) in str(result.exception)
            assert result.exit_code == 1

            matcher = EventMatcher(result.stderr)

            # tap/target pair
            assert matcher.event_matches("found ExtractLoadBlocks set")

            completed_events = matcher.find_by_event("Block run completed.")

            # there should only be one completed event
            assert len(completed_events) == 1
            assert completed_events[0]["success"] is False

            assert (
                completed_events[0]["err"]
                == "RunnerError('Output line length limit exceeded')"
            )

    @pytest.mark.backend("sqlite")
    @pytest.mark.usefixtures(
        "use_test_log_config",
        "project",
        "dbt",
        "dbt_process",
        "job_logging_service",
    )
    def test_run_mapper_config(
        self,
        cli_runner,
        tap,
        target,
        mapper,
        tap_process,
        target_process,
        mapper_process,
        project_add_service,
    ):
        # exit cleanly when everything is fine
        create_subprocess_exec = AsyncMock(
            side_effect=(tap_process, mapper_process, target_process),
        )

        # no mapper should be found
        args = ["run", tap.name, "not-a-valid-mapping-name", target.name]
        with mock.patch.object(SingerTap, "discover_catalog"), mock.patch.object(
            SingerTap,
            "apply_catalog_rules",
        ), mock.patch("meltano.core.plugin_invoker.asyncio") as asyncio_mock:
            asyncio_mock.create_subprocess_exec = create_subprocess_exec

            result = cli_runner.invoke(cli, args, catch_exceptions=True)
            assert result.exit_code == 1
            assert "Error: Block not-a-valid-mapping-name not found" in result.stderr

        # test mapper/mapping name collision detection - mapper plugin name no mappings
        project_add_service.add(
            PluginType.MAPPERS,
            "mapper-collision-01",
            inherit_from=mapper.name,
        )
        args = ["run", tap.name, "mapper-collision-01", target.name]
        with mock.patch.object(SingerTap, "discover_catalog"), mock.patch.object(
            SingerTap,
            "apply_catalog_rules",
        ), mock.patch("meltano.core.plugin_invoker.asyncio") as asyncio_mock2:
            asyncio_mock2.create_subprocess_exec = create_subprocess_exec
            with pytest.raises(
                Exception,
                match=(
                    "block violates set requirements: Expected unique mappings "
                    "name not the mapper plugin name: mapper-collision-01"
                ),
            ):
                cli_runner.invoke(cli, args, catch_exceptions=False)

        # Test mapper/mapping name collision detection - mappings name same a
        # mapper plugin name
        project_add_service.add(
            PluginType.MAPPERS,
            "mapper-collision-02",
            inherit_from=mapper.name,
            mappings=[
                {
                    "name": "mapper-collision-02",
                    "config": {
                        "transformations": [
                            {
                                "field_id": "author_email1",
                                "tap_stream_name": "commits1",
                                "type": "MASK-HIDDEN",
                            },
                        ],
                    },
                },
            ],
        )
        args = ["run", tap.name, "mapper-collision-02", target.name]
        with mock.patch.object(SingerTap, "discover_catalog"), mock.patch.object(
            SingerTap,
            "apply_catalog_rules",
        ), mock.patch("meltano.core.plugin_invoker.asyncio") as asyncio_mock2:
            asyncio_mock2.create_subprocess_exec = create_subprocess_exec
            with pytest.raises(
                Exception,
                match=(
                    "block violates set requirements: Expected unique mappings "
                    "name not the mapper plugin name: mapper-collision-02"
                ),
            ):
                cli_runner.invoke(cli, args, catch_exceptions=False)

        # create duplicate mapping name - should also fail
        project_add_service.add(
            PluginType.MAPPERS,
            "mapper-dupe1",
            inherit_from=mapper.name,
            mappings=[
                {
                    "name": "mock-mapping-dupe",
                    "config": {
                        "transformations": [
                            {
                                "field_id": "author_email1",
                                "tap_stream_name": "commits1",
                                "type": "MASK-HIDDEN",
                            },
                        ],
                    },
                },
            ],
        )
        project_add_service.add(
            PluginType.MAPPERS,
            "mapper-dupe2",
            inherit_from=mapper.name,
            mappings=[
                {
                    "name": "mock-mapping-dupe",
                    "config": {
                        "transformations": [
                            {
                                "field_id": "author_email1",
                                "tap_stream_name": "commits1",
                                "type": "MASK-HIDDEN",
                            },
                        ],
                    },
                },
            ],
        )

        args = ["run", tap.name, "mock-mapping-dupe", target.name]
        with mock.patch.object(SingerTap, "discover_catalog"), mock.patch.object(
            SingerTap,
            "apply_catalog_rules",
        ), mock.patch("meltano.core.plugin_invoker.asyncio") as asyncio_mock2:
            asyncio_mock2.create_subprocess_exec = create_subprocess_exec
            with pytest.raises(
                AmbiguousMappingName,
                match=(
                    "Ambiguous mapping name mock-mapping-dupe, found multiple matches."
                ),
            ):
                cli_runner.invoke(cli, args, catch_exceptions=False)

    @pytest.mark.backend("sqlite")
    @pytest.mark.usefixtures(
        "use_test_log_config",
        "project",
        "dbt",
        "job_logging_service",
    )
    def test_run_elb_mapper_failure(
        self,
        cli_runner,
        tap,
        target,
        mapper,
        tap_process,
        target_process,
        mapper_process,
        dbt_process,
    ):
        # In this scenario, the map fails on the second read. Target should
        # still complete, but dbt should not.
        args = ["run", tap.name, "mock-mapping-0", target.name, "dbt:run"]

        mapper_process.wait.return_value = 1
        mapper_process.returncode = 1
        mapper_process.stderr.readline.side_effect = (
            b"mapper starting\n",
            b"mapper running\n",
            b"mapper failure\n",
        )

        invoke_async = AsyncMock(
            side_effect=(tap_process, mapper_process, target_process, dbt_process),
        )

        with mock.patch.object(PluginInvoker, "invoke_async", new=invoke_async):
            result = cli_runner.invoke(cli, args, catch_exceptions=True)

            assert "Mappers failed" in str(result.exception)
            assert result.exit_code == 1

            matcher = EventMatcher(result.stderr)
            assert matcher.event_matches("found ExtractLoadBlocks set")
            assert (
                matcher.find_by_event("found PluginCommand")[0]["plugin_type"]
                == "transformers"
            )

            completed_events = matcher.find_by_event("Block run completed.")
            assert len(completed_events) == 1
            assert completed_events[0]["success"] is False

            # the tap should have completed successfully
            matcher.event_matches("tap done")

            # We should see a debug line for from the run manager indicating a
            # intermediate block failed
            matcher.event_matches("Intermediate block in sequence failed.")

            # the failed block should have been the mapper
            mapper_stop_event = matcher.find_by_event("mapper failure")
            assert len(mapper_stop_event) == 1
            assert mapper_stop_event[0]["name"] == mapper.name
            assert mapper_stop_event[0]["cmd_type"] == "elb"
            assert mapper_stop_event[0]["stdio"] == "stderr"

            # target should have attempted to complete
            target_stop_event = matcher.find_by_event("target done")
            assert len(target_stop_event) == 1
            assert target_stop_event[0]["name"] == target.name
            assert target_stop_event[0]["cmd_type"] == "elb"
            assert target_stop_event[0]["stdio"] == "stderr"

            # dbt should not have run at all
            assert not matcher.event_matches("dbt starting")
            assert not matcher.event_matches("dbt running")
            assert not matcher.event_matches("dbt done")

    @pytest.mark.backend("sqlite")
    @pytest.mark.usefixtures(
        "use_test_log_config",
        "project",
        "mapper",
        "dbt",
        "dbt_process",
        "job_logging_service",
    )
    def test_run_dry_run(
        self,
        cli_runner,
        tap,
        target,
        tap_process,
        target_process,
        mapper_process,
    ):
        # exit cleanly when everything is fine
        create_subprocess_exec = AsyncMock(
            side_effect=(tap_process, mapper_process, target_process),
        )

        args = ["run", "--dry-run", tap.name, "mock-mapping-0", target.name]
        with mock.patch.object(SingerTap, "discover_catalog"), mock.patch.object(
            SingerTap,
            "apply_catalog_rules",
        ), mock.patch("meltano.core.plugin_invoker.asyncio") as asyncio_mock:
            asyncio_mock.create_subprocess_exec = create_subprocess_exec
            result = cli_runner.invoke(cli, args, catch_exceptions=True)
            assert result.exit_code == 0

            matcher = EventMatcher(result.stderr)

            assert matcher.event_matches(
                "All ExtractLoadBlocks validated, starting execution.",
            )

            # We should have seen the --dry-run specific log line
            assert matcher.event_matches("Dry run, but would have run block 1/1.")

            # We should NOT see any mock done events, and definitely no block
            # completion log lines
            assert not matcher.find_by_event("tap done")
            assert not matcher.find_by_event("target done")
            assert not matcher.find_by_event("Block run completed.")
            assert create_subprocess_exec.call_count == 0
            assert asyncio_mock.call_count == 0

    @pytest.mark.backend("sqlite")
    @pytest.mark.usefixtures("project")
    @pytest.mark.parametrize("colors", (True, False))
    def test_color_console_exception_handler(
        self,
        colors,
        cli_runner,
        tap,
        target,
        tap_process,
        target_process,
        monkeypatch: pytest.MonkeyPatch,
    ):
        monkeypatch.delenv("FORCE_COLOR", raising=False)
        # toggle color in logging configuration
        logging_config = default_config(log_level="info")
        if not colors:
            logging_config["formatters"]["colored"] = {
                "()": "meltano.core.logging.console_log_formatter",
                "colors": colors,
            }

        # In this scenario, the tap fails on the third read. Target should
        # still complete.
        args = ["run", tap.name, target.name]

        tap_process.wait.return_value = 1
        tap_process.returncode = 1
        tap_process.stderr.readline.side_effect = (
            b"tap starting\n",
            b"tap running\n",
            b"tap failure\n",
        )

        invoke_async = AsyncMock(side_effect=(tap_process, target_process))

        with mock.patch(
            "meltano.core.logging.utils.default_config",
            return_value=logging_config,
        ), mock.patch.object(PluginInvoker, "invoke_async", new=invoke_async):
            result = cli_runner.invoke(cli, args)
            ansi_color_escape = re.compile(r"\x1b\[[0-9;]+m")
            match = ansi_color_escape.search(result.stderr)
            if colors:
                assert match
            else:
                assert not match


class TestUUIDParamType:
    @pytest.mark.parametrize(
        "value",
        (
            pytest.param("123e4567-e89b-12d3-a456-426614174000", id="with hyphens"),
            pytest.param("123e4567e89b12d3a456426614174000", id="without hyphens"),
        ),
    )
    def test_valid_uuid(self, value: str):
        param = UUIDParamType()
        assert param.convert(value, None, None) == uuid.UUID(value)

    def test_invalid_uuid(self):
        param = UUIDParamType()
        value = "zzz"
        with pytest.raises(click.BadParameter, match="is not a valid UUID"):
            param.convert(value, None, None)
